#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.nucaps module."""

import datetime
import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.float32
DEFAULT_FILE_SHAPE = (120,)
DEFAULT_PRES_FILE_SHAPE = (120, 100,)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_PRES_FILE_DATA = np.arange(DEFAULT_PRES_FILE_SHAPE[1], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_PRES_FILE_DATA = np.repeat([DEFAULT_PRES_FILE_DATA], DEFAULT_PRES_FILE_SHAPE[0], axis=0)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[0]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[0]).astype(DEFAULT_FILE_DTYPE)
ALL_PRESSURE_LEVELS = [
    0.0161, 0.0384, 0.0769, 0.137, 0.2244, 0.3454, 0.5064, 0.714,
    0.9753, 1.2972, 1.6872, 2.1526, 2.7009, 3.3398, 4.077, 4.9204,
    5.8776, 6.9567, 8.1655, 9.5119, 11.0038, 12.6492, 14.4559, 16.4318,
    18.5847, 20.9224, 23.4526, 26.1829, 29.121, 32.2744, 35.6505,
    39.2566, 43.1001, 47.1882, 51.5278, 56.126, 60.9895, 66.1253,
    71.5398, 77.2396, 83.231, 89.5204, 96.1138, 103.017, 110.237,
    117.777, 125.646, 133.846, 142.385, 151.266, 160.496, 170.078,
    180.018, 190.32, 200.989, 212.028, 223.441, 235.234, 247.408,
    259.969, 272.919, 286.262, 300, 314.137, 328.675, 343.618, 358.966,
    374.724, 390.893, 407.474, 424.47, 441.882, 459.712, 477.961,
    496.63, 515.72, 535.232, 555.167, 575.525, 596.306, 617.511, 639.14,
    661.192, 683.667, 706.565, 729.886, 753.628, 777.79, 802.371,
    827.371, 852.788, 878.62, 904.866, 931.524, 958.591, 986.067,
    1013.95, 1042.23, 1070.92, 1100
]
ALL_PRESSURE_LEVELS = np.repeat([ALL_PRESSURE_LEVELS], DEFAULT_PRES_FILE_SHAPE[0], axis=0)


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {
            "/attr/time_coverage_start": "2020-10-20T12:00:00.5Z",
            "/attr/time_coverage_end": "2020-10-20T12:00:36Z",
            "/attr/start_orbit_number": 1,
            "/attr/end_orbit_number": 2,
            "/attr/platform_name": "NPP",
            "/attr/instrument_name": "CrIS, ATMS, VIIRS",
        }
        for k, units, standard_name in [
            ("Solar_Zenith", "degrees", "solar_zenith_angle"),
            ("Topography", "meters", ""),
            ("Land_Fraction", "1", ""),
            ("Surface_Pressure", "mb", ""),
            ("Skin_Temperature", "Kelvin", "surface_temperature"),
        ]:
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
            file_content[k + "/attr/units"] = units
            file_content[k + "/attr/valid_range"] = (0., 120.)
            file_content[k + "/attr/_FillValue"] = -9999.
            if standard_name:
                file_content[k + "/attr/standard_name"] = standard_name
        for k, units, standard_name in [
            ("Temperature", "Kelvin", "air_temperature"),
            ("Effective_Pressure", "mb", ""),
            ("H2O", "1", ""),
            ("H2O_MR", "g/g", ""),
            ("O3", "1", ""),
            ("O3_MR", "1", ""),
            ("Liquid_H2O", "1", ""),
            ("Liquid_H2O_MR", "g/g", "cloud_liquid_water_mixing_ratio"),
            ("CO", "1", ""),
            ("CO_MR", "1", ""),
            ("CH4", "1", ""),
            ("CH4_MR", "1", ""),
            ("CO2", "1", ""),
            ("HNO3", "1", ""),
            ("HNO3_MR", "1", ""),
            ("N2O", "1", ""),
            ("N2O_MR", "1", ""),
            ("SO2", "1", ""),
            ("SO2_MR", "1", ""),
        ]:
            file_content[k] = DEFAULT_PRES_FILE_DATA
            file_content[k + "/shape"] = DEFAULT_PRES_FILE_SHAPE
            file_content[k + "/attr/units"] = units
            file_content[k + "/attr/valid_range"] = (0., 120.)
            file_content[k + "/attr/_FillValue"] = -9999.
            if standard_name:
                file_content[k + "/attr/standard_name"] = standard_name
        k = "Pressure"
        file_content[k] = ALL_PRESSURE_LEVELS
        file_content[k + "/shape"] = DEFAULT_PRES_FILE_SHAPE
        file_content[k + "/attr/units"] = "mb"
        file_content[k + "/attr/valid_range"] = (0., 2000.)
        file_content[k + "/attr/_FillValue"] = -9999.

        k = "Quality_Flag"
        file_content[k] = DEFAULT_FILE_DATA.astype(np.int32)
        file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
        file_content[k + "/attr/valid_range"] = (0, 31)
        file_content[k + "/attr/_FillValue"] = -9999.

        k = "Longitude"
        file_content[k] = DEFAULT_LON_DATA
        file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
        file_content[k + "/attr/units"] = "degrees_east"
        file_content[k + "/attr/valid_range"] = (-180., 180.)
        file_content[k + "/attr/standard_name"] = "longitude"
        file_content[k + "/attr/_FillValue"] = -9999.

        k = "Latitude"
        file_content[k] = DEFAULT_LAT_DATA
        file_content[k + "/shape"] = DEFAULT_FILE_SHAPE
        file_content[k + "/attr/units"] = "degrees_north"
        file_content[k + "/attr/valid_range"] = (-90., 90.)
        file_content[k + "/attr/standard_name"] = "latitude"
        file_content[k + "/attr/_FillValue"] = -9999.

        attrs = ("_FillValue", "flag_meanings", "flag_values", "units")
        cris_fors_dim_name = "Number_of_CrIS_FORs"
        pressure_levels_dim_name = "Number_of_P_Levels"
        if ("_v1" in filename):
            cris_fors_dim_name = "number_of_FORs"
            pressure_levels_dim_name = "number_of_p_levels"
        convert_file_content_to_data_array(
            file_content, attrs=attrs,
            dims=("z", cris_fors_dim_name, pressure_levels_dim_name))
        return file_content


class TestNUCAPSReader(unittest.TestCase):
    """Test NUCAPS Reader."""

    yaml_file = "nucaps.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.nucaps import NUCAPSFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(NUCAPSFileHandler, "__bases__", (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_init_with_kwargs(self):
        """Test basic init with extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs, mask_surface=False)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables, fh_kwargs={"mask_surface": False})
        # make sure we have some files
        assert r.file_handlers

    def test_load_nonpressure_based(self):
        """Test loading all channels that aren't based on pressure."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Solar_Zenith",
                           "Topography",
                           "Land_Fraction",
                           "Surface_Pressure",
                           "Skin_Temperature",
                           "Quality_Flag",
                           ])
        assert len(datasets) == 6
        for v in datasets.values():
            # self.assertNotEqual(v.info['resolution'], 0)
            # self.assertEqual(v.info['units'], 'degrees')
            assert v.ndim == 1
            assert v.attrs["sensor"] == set(["cris", "atms", "viirs"])
            assert isinstance(v.attrs["start_time"], datetime.datetime)
            assert isinstance(v.attrs["end_time"], datetime.datetime)

    def test_load_pressure_based(self):
        """Test loading all channels based on pressure."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature",
                           "Effective_Pressure",
                           "H2O",
                           "H2O_MR",
                           "O3",
                           "O3_MR",
                           "Liquid_H2O",
                           "Liquid_H2O_MR",
                           "CO",
                           "CO_MR",
                           "CH4",
                           "CH4_MR",
                           "CO2",
                           "HNO3",
                           "HNO3_MR",
                           "N2O",
                           "N2O_MR",
                           "SO2",
                           "SO2_MR",
                           ])
        assert len(datasets) == 19
        for v in datasets.values():
            # self.assertNotEqual(v.info['resolution'], 0)
            assert v.ndim == 2
            if np.issubdtype(v.dtype, np.floating):
                assert "_FillValue" not in v.attrs

    def test_load_multiple_files_pressure(self):
        """Test loading Temperature from multiple input files."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
            "NUCAPS-EDR_v1r0_npp_s201603011159009_e201603011159307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=True)
        assert len(datasets) == 100
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_individual_pressure_levels_true(self):
        """Test loading Temperature with individual pressure datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=True)
        assert len(datasets) == 100
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_individual_pressure_levels_min_max(self):
        """Test loading individual Temperature with min/max level specified."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=(100., 150.))
        assert len(datasets) == 6
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_individual_pressure_levels_single(self):
        """Test loading individual Temperature with specific levels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=(103.017,))
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_pressure_levels_true(self):
        """Test loading Temperature with all pressure levels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature"], pressure_levels=True)
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 2
            assert v.shape == DEFAULT_PRES_FILE_SHAPE

    def test_load_pressure_levels_min_max(self):
        """Test loading Temperature with min/max level specified."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature"], pressure_levels=(100., 150.))
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 2
            assert v.shape == (DEFAULT_PRES_FILE_SHAPE[0], 6)

    def test_load_pressure_levels_single(self):
        """Test loading a specific Temperature level."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature"], pressure_levels=(103.017,))
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 2
            assert v.shape == (DEFAULT_PRES_FILE_SHAPE[0], 1)

    def test_load_pressure_levels_single_and_pressure_levels(self):
        """Test loading a specific Temperature level and pressure levels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature", "Pressure_Levels"], pressure_levels=(103.017,))
        assert len(datasets) == 2
        t_ds = datasets["Temperature"]
        assert t_ds.ndim == 2
        assert t_ds.shape == (DEFAULT_PRES_FILE_SHAPE[0], 1)
        pl_ds = datasets["Pressure_Levels"]
        assert pl_ds.shape == (1,)


class TestNUCAPSScienceEDRReader(unittest.TestCase):
    """Test NUCAPS Science EDR Reader."""

    yaml_file = "nucaps.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.nucaps import NUCAPSFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(NUCAPSFileHandler, "__bases__", (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_nonpressure_based(self):
        """Test loading all channels that aren't based on pressure."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Topography",
                           "Land_Fraction",
                           "Surface_Pressure",
                           "Skin_Temperature",
                           "Quality_Flag",
                           ])
        assert len(datasets) == 5
        for v in datasets.values():
            assert v.ndim == 1
            assert v.attrs["sensor"] == set(["cris", "atms", "viirs"])
            assert isinstance(v.attrs["start_time"], datetime.datetime)
            assert isinstance(v.attrs["end_time"], datetime.datetime)

    def test_load_pressure_based(self):
        """Test loading all channels based on pressure."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature",
                           "H2O",
                           "H2O_MR",
                           "O3",
                           "O3_MR",
                           "CO",
                           "CO_MR",
                           "CH4",
                           "CH4_MR",
                           "CO2",
                           "HNO3",
                           "HNO3_MR",
                           "N2O",
                           "N2O_MR",
                           "SO2",
                           "SO2_MR",
                           ])
        assert len(datasets) == 16
        for v in datasets.values():
            # self.assertNotEqual(v.info['resolution'], 0)
            assert v.ndim == 2

    def test_load_individual_pressure_levels_true(self):
        """Test loading Temperature with individual pressure datasets."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=True)
        assert len(datasets) == 100
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_individual_pressure_levels_min_max(self):
        """Test loading individual Temperature with min/max level specified."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=(100., 150.))
        assert len(datasets) == 6
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_individual_pressure_levels_single(self):
        """Test loading individual Temperature with specific levels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names["Temperature"], pressure_levels=(103.017,))
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 1

    def test_load_pressure_levels_true(self):
        """Test loading Temperature with all pressure levels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature"], pressure_levels=True)
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 2
            assert v.shape == DEFAULT_PRES_FILE_SHAPE

    def test_load_pressure_levels_min_max(self):
        """Test loading Temperature with min/max level specified."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature"], pressure_levels=(100., 150.))
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 2
            assert v.shape == (DEFAULT_PRES_FILE_SHAPE[0], 6)

    def test_load_pressure_levels_single(self):
        """Test loading a specific Temperature level."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature"], pressure_levels=(103.017,))
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.ndim == 2
            assert v.shape == (DEFAULT_PRES_FILE_SHAPE[0], 1)

    def test_load_pressure_levels_single_and_pressure_levels(self):
        """Test loading a specific Temperature level and pressure levels."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc",
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["Temperature", "Pressure_Levels"], pressure_levels=(103.017,))
        assert len(datasets) == 2
        t_ds = datasets["Temperature"]
        assert t_ds.ndim == 2
        assert t_ds.shape == (DEFAULT_PRES_FILE_SHAPE[0], 1)
        pl_ds = datasets["Pressure_Levels"]
        assert pl_ds.shape == (1,)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the satpy.readers.tropomi_l2 module."""

import datetime as dt
import os
import unittest
from unittest import mock

import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (3246, 450)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_BOUND_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1] * 4,
                               dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE+(4,))


class FakeNetCDF4FileHandlerTL2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        dt_s = filename_info.get("start_time", dt.datetime(2016, 1, 1, 12, 0, 0))
        dt_e = filename_info.get("end_time", dt.datetime(2016, 1, 1, 12, 0, 0))

        if filetype_info["file_type"] == "tropomi_l2":
            file_content = {
                "/attr/time_coverage_start": (dt_s+dt.timedelta(minutes=22)).strftime("%Y-%m-%dT%H:%M:%SZ"),
                "/attr/time_coverage_end": (dt_e-dt.timedelta(minutes=22)).strftime("%Y-%m-%dT%H:%M:%SZ"),
                "/attr/platform_shortname": "S5P",
                "/attr/sensor": "TROPOMI",
            }

            file_content["PRODUCT/latitude"] = DEFAULT_FILE_DATA
            file_content["PRODUCT/longitude"] = DEFAULT_FILE_DATA
            file_content["PRODUCT/SUPPORT_DATA/GEOLOCATIONS/latitude_bounds"] = DEFAULT_BOUND_DATA
            file_content["PRODUCT/SUPPORT_DATA/GEOLOCATIONS/longitude_bounds"] = DEFAULT_BOUND_DATA

            if "NO2" in filename:
                file_content["PRODUCT/nitrogen_dioxide_total_column"] = DEFAULT_FILE_DATA
            if "SO2" in filename:
                file_content["PRODUCT/sulfurdioxide_total_vertical_column"] = DEFAULT_FILE_DATA

            for k in list(file_content.keys()):
                if not k.startswith("PRODUCT"):
                    continue
                file_content[k + "/shape"] = DEFAULT_FILE_SHAPE

            self._convert_data_content_to_dataarrays(file_content)
            file_content["PRODUCT/latitude"].attrs["_FillValue"] = -999.0
            file_content["PRODUCT/longitude"].attrs["_FillValue"] = -999.0
            file_content["PRODUCT/SUPPORT_DATA/GEOLOCATIONS/latitude_bounds"].attrs["_FillValue"] = -999.0
            file_content["PRODUCT/SUPPORT_DATA/GEOLOCATIONS/longitude_bounds"].attrs["_FillValue"] = -999.0
            if "NO2" in filename:
                file_content["PRODUCT/nitrogen_dioxide_total_column"].attrs["_FillValue"] = -999.0
            if "SO2" in filename:
                file_content["PRODUCT/sulfurdioxide_total_vertical_column"].attrs["_FillValue"] = -999.0

        else:
            raise NotImplementedError("Test data for file types other than "
                                      "'tropomi_l2' are not supported.")

        return file_content

    def _convert_data_content_to_dataarrays(self, file_content):
        """Convert data content to xarray's dataarrays."""
        from xarray import DataArray
        for key, val in file_content.items():
            if isinstance(val, np.ndarray):
                if 1 < val.ndim <= 2:
                    file_content[key] = DataArray(val, dims=("scanline", "ground_pixel"))
                elif val.ndim > 2:
                    file_content[key] = DataArray(val, dims=("scanline", "ground_pixel", "corner"))
                else:
                    file_content[key] = DataArray(val)


class TestTROPOMIL2Reader(unittest.TestCase):
    """Test TROPOMI L2 Reader."""

    yaml_file = "tropomi_l2.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.tropomi_l2 import TROPOMIL2FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(TROPOMIL2FileHandler, "__bases__", (FakeNetCDF4FileHandlerTL2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic initialization of this reader."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "S5P_OFFL_L2__NO2____20180709T170334_20180709T184504_03821_01_010002_20180715T184729.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_load_no2(self):
        """Load NO2 dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.tropomi_l2.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "S5P_OFFL_L2__NO2____20180709T170334_20180709T184504_03821_01_010002_20180715T184729.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(["nitrogen_dioxide_total_column"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "S5P"
            assert d.attrs["sensor"] == "tropomi"
            assert d.attrs["time_coverage_start"] == dt.datetime(2018, 7, 9, 17, 25, 34)
            assert d.attrs["time_coverage_end"] == dt.datetime(2018, 7, 9, 18, 23, 4)
            assert "area" in d.attrs
            assert d.attrs["area"] is not None
            assert "y" in d.dims
            assert "x" in d.dims

    def test_load_so2(self):
        """Load SO2 dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.tropomi_l2.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "S5P_OFFL_L2__SO2____20181224T055107_20181224T073237_06198_01_010105_20181230T150634.nc",
            ])
            r.create_filehandlers(loadables)
        ds = r.load(["sulfurdioxide_total_vertical_column"])
        assert len(ds) == 1
        for d in ds.values():
            assert d.attrs["platform_shortname"] == "S5P"
            assert "area" in d.attrs
            assert d.attrs["area"] is not None
            assert "y" in d.dims
            assert "x" in d.dims

    def test_load_bounds(self):
        """Load bounds dataset."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.tropomi_l2.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "S5P_OFFL_L2__NO2____20180709T170334_20180709T184504_03821_01_010002_20180715T184729.nc",
            ])
            r.create_filehandlers(loadables)
        keys = ["latitude_bounds", "longitude_bounds"]
        ds = r.load(keys)
        assert len(ds) == 2
        for key in keys:
            assert ds[key].attrs["platform_shortname"] == "S5P"
            assert "y" in ds[key].dims
            assert "x" in ds[key].dims
            assert "corner" in ds[key].dims
            # check assembled bounds
            left = np.vstack([ds[key][:, :, 0], ds[key][-1:, :, 3]])
            right = np.vstack([ds[key][:, -1:, 1], ds[key][-1:, -1:, 2]])
            dest = np.hstack([left, right])
            dest = xr.DataArray(dest,
                                dims=("y", "x")
                                )
            dest.attrs = ds[key].attrs
            assert dest.attrs["platform_shortname"] == "S5P"
            assert "y" in dest.dims
            assert "x" in dest.dims
            assert DEFAULT_FILE_SHAPE[0] + 1 == dest.shape[0]
            assert DEFAULT_FILE_SHAPE[1] + 1 == dest.shape[1]
            np.testing.assert_array_equal(dest[:-1, :-1], ds[key][:, :, 0])
            np.testing.assert_array_equal(dest[-1, :-1], ds[key][-1, :, 3])
            np.testing.assert_array_equal(dest[:, -1], np.append(ds[key][:, -1, 1], ds[key][-1:, -1:, 2]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for the VIIRS instrument."""

import logging

import numpy as np

from satpy.composites.core import GenericCompositor
from satpy.dataset import combine_metadata

LOG = logging.getLogger(__name__)


def overlay(top, bottom, maxval=None):
    """Blending two layers.

    from: https://docs.gimp.org/en/gimp-concepts-layer-modes.html
    """
    if maxval is None:
        maxval = np.maximum(top.max(), bottom.max())

    res = ((2 * top / maxval - 1) * bottom + 2 * top) * bottom / maxval
    return res.clip(min=0)


def soft_light(top, bottom, maxval):
    """Apply soft light.

    http://www.pegtop.net/delphi/articles/blendmodes/softlight.htm
    """
    a = top / maxval
    b = bottom / maxval
    return (2*a*b + a*a * (1 - 2*b)) * maxval


class SARIce(GenericCompositor):
    """The SAR Ice composite."""

    def __call__(self, projectables, *args, **kwargs):
        """Create the SAR Ice composite."""
        (mhh, mhv) = projectables
        ch1attrs = mhh.attrs
        ch2attrs = mhv.attrs
        mhh = np.sqrt(mhh + 0.002) - 0.04
        mhv = np.sqrt(mhv + 0.002) - 0.04
        mhh.attrs = ch1attrs
        mhv.attrs = ch2attrs
        green = overlay(mhh, mhv, 30) * 1000
        green.attrs = combine_metadata(mhh, mhv)

        return super(SARIce, self).__call__((mhv, green, mhh), *args, **kwargs)


def _square_root_channels(*projectables):
    """Return the square root of the channels, preserving the attributes."""
    results = []
    for projectable in projectables:
        attrs = projectable.attrs
        projectable = np.sqrt(projectable)
        projectable.attrs = attrs
        results.append(projectable)
    return results


class SARIceLegacy(GenericCompositor):
    """The SAR Ice composite, legacy version with dynamic stretching."""

    def __call__(self, projectables, *args, **kwargs):
        """Create the SAR RGB composite."""
        mhh, mhv = _square_root_channels(*projectables)
        green = overlay(mhh, mhv)
        green.attrs = combine_metadata(mhh, mhv)

        return super(SARIceLegacy, self).__call__((mhv, green, mhh), *args, **kwargs)


class SARIceLog(GenericCompositor):
    """The SAR Ice composite, using log-scale data."""

    def __call__(self, projectables, *args, **kwargs):
        """Create the SAR Ice Log composite."""
        mhh, mhv = projectables
        mhh = mhh.clip(-40)
        mhv = mhv.clip(-38)
        green = soft_light(mhh + 100, mhv + 100, 100) - 100
        green.attrs = combine_metadata(mhh, mhv)

        return super().__call__((mhv, green, mhh), *args, **kwargs)


class SARRGB(GenericCompositor):
    """The SAR RGB composite."""

    def __call__(self, projectables, *args, **kwargs):
        """Create the SAR RGB composite."""
        mhh, mhv = _square_root_channels(*projectables)
        green = overlay(mhh, mhv)
        green.attrs = combine_metadata(mhh, mhv)

        return super(SARRGB, self).__call__((-mhv, -green, -mhh), *args, **kwargs)


class SARQuickLook(GenericCompositor):
    """The SAR QuickLook composite."""

    def __call__(self, projectables, *args, **kwargs):
        """Create the SAR QuickLook composite."""
        mhh, mhv = _square_root_channels(*projectables)

        blue = mhv / mhh
        blue.attrs = combine_metadata(mhh, mhv)

        return super(SARQuickLook, self).__call__((mhh, mhv, blue), *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for GERB L2 HR HDF5 reader."""

import h5py
import numpy as np
import pytest

from satpy import Scene

FNAME = "G4_SEV4_L20_HR_SOL_TH_20190606_130000_V000.hdf"


def make_h5_null_string(length):
    """Make a HDF5 type for a NULL terminated string of fixed length."""
    dt = h5py.h5t.TypeID.copy(h5py.h5t.C_S1)
    dt.set_size(7)
    dt.set_strpad(h5py.h5t.STR_NULLTERM)
    return dt


def write_h5_null_string_att(loc_id, name, s):
    """Write a NULL terminated string attribute at loc_id."""
    dt = make_h5_null_string(length=7)
    name = bytes(name.encode("ascii"))
    s = bytes(s.encode("ascii"))
    at = h5py.h5a.create(loc_id, name, dt, h5py.h5s.create(h5py.h5s.SCALAR))
    at.write(np.array(s, dtype=f"|S{len(s)+1}"))


@pytest.fixture(scope="session")
def gerb_l2_hr_h5_dummy_file(tmp_path_factory):
    """Create a dummy HDF5 file for the GERB L2 HR product."""
    filename = tmp_path_factory.mktemp("data") / FNAME

    with h5py.File(filename, "w") as fid:
        fid.create_group("/Angles")
        fid["/Angles/Relative Azimuth"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Angles/Relative Azimuth"].attrs["Quantisation Factor"] = np.array(0.1, dtype="float64")
        fid["/Angles/Solar Zenith"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Angles/Solar Zenith"].attrs["Quantisation Factor"] = np.array(0.1, dtype="float64")
        write_h5_null_string_att(fid["/Angles/Relative Azimuth"].id, "Unit", "Degree")
        fid["/Angles/Viewing Azimuth"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Angles/Viewing Azimuth"].attrs["Quantisation Factor"] = np.array(0.1, dtype="float64")
        write_h5_null_string_att(fid["/Angles/Viewing Azimuth"].id, "Unit", "Degree")
        fid["/Angles/Viewing Zenith"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Angles/Viewing Zenith"].attrs["Quantisation Factor"] = np.array(0.1, dtype="float64")
        write_h5_null_string_att(fid["/Angles/Viewing Zenith"].id, "Unit", "Degree")
        fid.create_group("/GERB")
        dt = h5py.h5t.TypeID.copy(h5py.h5t.C_S1)
        dt.set_size(3)
        dt.set_strpad(h5py.h5t.STR_NULLTERM)
        write_h5_null_string_att(fid["/GERB"].id, "Instrument Identifier", "G4")
        fid.create_group("/GGSPS")
        fid["/GGSPS"].attrs["L1.5 NANRG Product Version"] = np.array(-1, dtype="int32")
        fid.create_group("/Geolocation")
        write_h5_null_string_att(fid["/Geolocation"].id, "Geolocation File Name",
                                 "G4_SEV4_L20_HR_GEO_20180111_181500_V010.hdf")
        fid["/Geolocation"].attrs["Nominal Satellite Longitude (degrees)"] = np.array(0.0, dtype="float64")
        fid.create_group("/Imager")
        fid["/Imager"].attrs["Instrument Identifier"] = np.array(4, dtype="int32")
        write_h5_null_string_att(fid["/Imager"].id, "Type", "SEVIRI")
        fid.create_group("/RMIB")
        fid.create_group("/Radiometry")
        fid["/Radiometry"].attrs["SEVIRI Radiance Definition Flag"] = np.array(2, dtype="int32")
        fid["/Radiometry/A Values (per GERB detector cell)"] = np.ones(shape=(256,), dtype=np.dtype(">f8"))
        fid["/Radiometry/C Values (per GERB detector cell)"] = np.ones(shape=(256,), dtype=np.dtype(">f8"))
        fid["/Radiometry/Longwave Correction"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Radiometry/Longwave Correction"].attrs["Offset"] = np.array(1.0, dtype="float64")
        fid["/Radiometry/Longwave Correction"].attrs["Quantisation Factor"] = np.array(0.005, dtype="float64")
        fid["/Radiometry/Shortwave Correction"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Radiometry/Shortwave Correction"].attrs["Offset"] = np.array(1.0, dtype="float64")
        fid["/Radiometry/Shortwave Correction"].attrs["Quantisation Factor"] = np.array(0.005, dtype="float64")
        fid["/Radiometry/Solar Flux"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Radiometry/Solar Flux"].attrs["Quantisation Factor"] = np.array(0.25, dtype="float64")
        write_h5_null_string_att(fid["/Radiometry/Solar Flux"].id, "Unit", "Watt per square meter")
        fid["/Radiometry/Solar Radiance"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Radiometry/Solar Radiance"].attrs["Quantisation Factor"] = np.array(0.05, dtype="float64")
        write_h5_null_string_att(fid["/Radiometry/Solar Radiance"].id, "Unit", "Watt per square meter per steradian")
        fid["/Radiometry/Thermal Flux"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Radiometry/Thermal Flux"].attrs["Quantisation Factor"] = np.array(0.25, dtype="float64")
        write_h5_null_string_att(fid["/Radiometry/Thermal Flux"].id, "Unit", "Watt per square meter")
        fid["/Radiometry/Thermal Radiance"] = np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Radiometry/Thermal Radiance"].attrs["Quantisation Factor"] = np.array(0.05, dtype="float64")
        write_h5_null_string_att(fid["/Radiometry/Thermal Radiance"].id, "Unit", "Watt per square meter per steradian")
        fid.create_group("/Scene Identification")
        write_h5_null_string_att(fid["/Scene Identification"].id,
                                 "Solar Angular Dependency Models Set Version", "CERES_TRMM.1")
        write_h5_null_string_att(fid["/Scene Identification"].id,
                                 "Thermal Angular Dependency Models Set Version", "RMIB.3")
        fid["/Scene Identification/Cloud Cover"] = np.ones(shape=(1237, 1237), dtype=np.dtype("uint8"))
        fid["/Scene Identification/Cloud Cover"].attrs["Quantisation Factor"] = np.array(0.01, dtype="float64")
        write_h5_null_string_att(fid["/Scene Identification/Cloud Cover"].id, "Unit", "Percent")
        fid["/Scene Identification/Cloud Optical Depth (logarithm)"] = \
            np.ones(shape=(1237, 1237), dtype=np.dtype(">i2"))
        fid["/Scene Identification/Cloud Optical Depth (logarithm)"].attrs["Quantisation Factor"] = \
            np.array(0.00025, dtype="float64")
        fid["/Scene Identification/Cloud Phase"] = np.ones(shape=(1237, 1237), dtype=np.dtype("uint8"))
        fid["/Scene Identification/Cloud Phase"].attrs["Quantisation Factor"] = np.array(0.01, dtype="float64")
        write_h5_null_string_att(fid["/Scene Identification/Cloud Phase"].id, "Unit",
                                 "Percent (Water=0%,Mixed,Ice=100%)")
        fid.create_group("/Times")
        fid["/Times/Time (per row)"] = np.ones(shape=(1237,), dtype=np.dtype("|S22"))

    return filename


@pytest.mark.parametrize("name", ["Solar Flux", "Thermal Flux", "Solar Radiance", "Thermal Radiance"])
@pytest.mark.parametrize("area", [None, "msg_seviri_iodc_9km"])
def test_dataset_load(gerb_l2_hr_h5_dummy_file, name, area):
    """Test loading the solar flux component."""
    scene = Scene(reader="gerb_l2_hr_h5", filenames=[gerb_l2_hr_h5_dummy_file],
                  reader_kwargs={"area": area})
    scene.load([name])
    assert scene[name].shape == (1237, 1237)
    assert np.nanmax((scene[name].to_numpy().flatten() - 0.25)) < 1e-6



# Copyright (c) 2009-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module defining various utilities."""

from __future__ import annotations

import contextlib
import datetime
import importlib.metadata
import logging
import os
import pathlib
import platform
import warnings
from collections.abc import Mapping, MutableMapping
from contextlib import contextmanager
from copy import deepcopy
from typing import Any, Literal, Optional
from urllib.parse import urlparse

import dask.utils
import numpy as np
import xarray as xr
import yaml
from numpy.typing import ArrayLike, DTypeLike
from yaml import BaseLoader, UnsafeLoader

_is_logging_on = False
TRACE_LEVEL = 5

logger = logging.getLogger(__name__)


class PerformanceWarning(Warning):
    """Warning raised when there is a possible performance impact."""


def debug_on(deprecation_warnings=True):
    """Turn debugging logging on.

    Sets up a StreamHandler to to `sys.stderr` at debug level for all
    loggers, such that all debug messages (and log messages with higher
    severity) are logged to the standard error stream.

    By default, since Satpy 0.26, this also enables the global visibility
    of deprecation warnings.  This can be suppressed by passing a false
    value.

    Args:
        deprecation_warnings (Optional[bool]): Switch on deprecation warnings.
            Defaults to True.

    Returns:
        None
    """
    logging_on(logging.DEBUG)
    if deprecation_warnings:
        deprecation_warnings_on()


def debug_off():
    """Turn debugging logging off.

    This disables both debugging logging and the global visibility of
    deprecation warnings.
    """
    logging_off()
    deprecation_warnings_off()


@contextlib.contextmanager
def debug(deprecation_warnings=True):
    """Context manager to temporarily set debugging on.

    Example::

        >>> with satpy.utils.debug():
        ...     code_here()

    Args:
        deprecation_warnings (Optional[bool]): Switch on deprecation warnings.
            Defaults to True.
    """
    debug_on(deprecation_warnings=deprecation_warnings)
    yield
    debug_off()


def trace_on():
    """Turn trace logging on."""
    logging_on(TRACE_LEVEL)


class _WarningManager:
    """Class to handle switching warnings on and off."""

    filt = None


_warning_manager = _WarningManager()


def deprecation_warnings_on():
    """Switch on deprecation warnings."""
    warnings.filterwarnings("default", category=DeprecationWarning)
    _warning_manager.filt = warnings.filters[0]


def deprecation_warnings_off():
    """Switch off deprecation warnings."""
    if _warning_manager.filt in warnings.filters:
        warnings.filters.remove(_warning_manager.filt)


def logging_on(level=logging.WARNING):
    """Turn logging on."""
    global _is_logging_on

    if not _is_logging_on:
        console = logging.StreamHandler()
        console.setFormatter(logging.Formatter("[%(levelname)s: %(asctime)s :"
                                               " %(name)s] %(message)s",
                                               "%Y-%m-%d %H:%M:%S"))
        console.setLevel(level)
        logging.getLogger("").addHandler(console)
        _is_logging_on = True

    log = logging.getLogger("")
    log.setLevel(level)
    for h in log.handlers:
        h.setLevel(level)


def logging_off():
    """Turn logging off."""
    logging.getLogger("").handlers = [logging.NullHandler()]


def get_logger(name):
    """Return logger with null handler added if needed."""
    if not hasattr(logging.Logger, "trace"):
        logging.addLevelName(TRACE_LEVEL, "TRACE")

        def trace(self, message, *args, **kwargs):
            if self.isEnabledFor(TRACE_LEVEL):
                # Yes, logger takes its '*args' as 'args'.
                self._log(TRACE_LEVEL, message, args, **kwargs)

        logging.Logger.trace = trace

    log = logging.getLogger(name)
    return log


def in_ipynb():
    """Check if we are in a jupyter notebook."""
    try:
        return "ZMQ" in get_ipython().__class__.__name__
    except NameError:
        return False


# Spherical conversions


def lonlat2xyz(
        lon: ArrayLike,
        lat: ArrayLike,
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    """Convert lon lat to cartesian.

    For a sphere with unit radius, convert the spherical coordinates
    longitude and latitude to cartesian coordinates.

    Args:
        lon: Longitude in °.
        lat: Latitude in °.

    Returns:
        (x, y, z) Cartesian coordinates [1]
    """
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)
    x = np.cos(lat) * np.cos(lon)
    y = np.cos(lat) * np.sin(lon)
    z = np.sin(lat)
    return x, y, z


def xyz2lonlat(
        x: ArrayLike,
        y: ArrayLike,
        z: ArrayLike,
        asin: bool = False,
) -> tuple[ArrayLike, ArrayLike]:
    """Convert cartesian to lon lat.

    For a sphere with unit radius, convert cartesian coordinates to spherical
    coordinates longitude and latitude.

    Args:
        x: x-coordinate, unitless
        y: y-coordinate, unitless
        z: z-coordinate, unitless
        asin: If true, use arcsin for calculations.
            If false, use arctan2 for calculations.

    Returns:
        Longitude and latitude in °.
    """
    lon = np.rad2deg(np.arctan2(y, x))
    if asin:
        lat = np.rad2deg(np.arcsin(z))
    else:
        lat = np.rad2deg(np.arctan2(z, np.sqrt(x ** 2 + y ** 2)))
    return lon, lat


def angle2xyz(azi, zen):
    """Convert azimuth and zenith to cartesian."""
    azi = np.deg2rad(azi)
    zen = np.deg2rad(zen)
    x = np.sin(zen) * np.sin(azi)
    y = np.sin(zen) * np.cos(azi)
    z = np.cos(zen)
    return x, y, z


def xyz2angle(x, y, z, acos=False):
    """Convert cartesian to azimuth and zenith."""
    azi = np.rad2deg(np.arctan2(x, y))
    if acos:
        zen = np.rad2deg(np.arccos(z))
    else:
        zen = 90 - np.rad2deg(np.arctan2(z, np.sqrt(x ** 2 + y ** 2)))
    return azi, zen


def proj_units_to_meters(proj_str):
    """Convert projection units from kilometers to meters."""
    proj_parts = proj_str.split()
    new_parts = []
    for itm in proj_parts:
        key, val = itm.split("=")
        key = key.strip("+")
        if key in ["a", "b", "h"]:
            val = float(val)
            if val < 6e6:
                val *= 1000.
                val = "%.3f" % val

        if key == "units" and val == "km":
            continue

        new_parts.append("+%s=%s" % (key, val))

    return " ".join(new_parts)


def _get_sunz_corr_li_and_shibata(cos_zen):
    return 24.35 / (2. * cos_zen + np.sqrt(498.5225 * cos_zen**2 + 1))


def atmospheric_path_length_correction(data, cos_zen, limit=88., max_sza=95.):
    """Perform Sun zenith angle correction.

    This function uses the correction method proposed by
    Li and Shibata (2006): https://doi.org/10.1175/JAS3682.1

    The correction is limited to ``limit`` degrees (default: 88.0 degrees). For
    larger zenith angles, the correction is the same as at the ``limit`` if
    ``max_sza`` is `None`. The default behavior is to gradually reduce the
    correction past ``limit`` degrees up to ``max_sza`` where the correction
    becomes 0. Both ``data`` and ``cos_zen`` should be 2D arrays of the same
    shape.

    """
    # Convert the zenith angle limit to cosine of zenith angle
    limit_rad = np.deg2rad(limit)
    limit_cos = np.cos(limit_rad)
    max_sza_rad = np.deg2rad(max_sza) if max_sza is not None else max_sza

    # Cosine correction
    corr = _get_sunz_corr_li_and_shibata(cos_zen)
    # Use constant value (the limit) for larger zenith angles
    corr_lim = _get_sunz_corr_li_and_shibata(limit_cos)

    if max_sza is not None:
        # gradually fall off for larger zenith angle
        grad_factor = (np.arccos(cos_zen) - limit_rad) / (max_sza_rad - limit_rad)
        # invert the factor so maximum correction is done at `limit` and falls off later
        grad_factor = 1. - np.log(grad_factor + 1) / np.log(2)
        # make sure we don't make anything negative
        grad_factor = grad_factor.clip(0.)
    else:
        # Use constant value (the limit) for larger zenith angles
        grad_factor = 1.
    corr = corr.where(cos_zen > limit_cos, grad_factor * corr_lim)
    # Force "night" pixels to 0 (where SZA is invalid)
    corr = corr.where(cos_zen.notnull(), 0)

    return data * corr


def get_satpos(
        data_arr: xr.DataArray,
        preference: Optional[str] = None,
        use_tle: bool = False
) -> tuple[float, float, float]:
    """Get satellite position from dataset attributes.

    Args:
        data_arr: DataArray object to access ``.attrs`` metadata
            from.
        preference: Optional preference for one of the available types of
            position information. If not provided or ``None`` then the default
            preference is:

            * Longitude & Latitude: nadir, actual, nominal, projection
            * Altitude: actual, nominal, projection

            The provided ``preference`` can be any one of these individual
            strings (nadir, actual, nominal, projection). If the
            preference is not available then the original preference list is
            used. A warning is issued when projection values have to be used because
            nothing else is available and it wasn't provided as the ``preference``.
        use_tle: If true, try to obtain position via satellite name
            and TLE if it can't be determined otherwise.  This requires pyorbital, skyfield,
            and astropy to be installed and may need network access to obtain the TLE.
            Note that even if ``use_tle`` is true, the TLE will not be used if
            the dataset metadata contain the satellite position directly.

    Returns:
        Geodetic longitude, latitude, altitude [m]

    """
    if preference is not None and preference not in ("nadir", "actual", "nominal", "projection"):
        raise ValueError(f"Unrecognized satellite coordinate preference: {preference}")
    lonlat_prefixes = ("nadir_", "satellite_actual_", "satellite_nominal_", "projection_")
    alt_prefixes = _get_prefix_order_by_preference(lonlat_prefixes[1:], preference)
    lonlat_prefixes = _get_prefix_order_by_preference(lonlat_prefixes, preference)
    try:
        lon, lat = _get_sat_lonlat(data_arr, lonlat_prefixes)
        alt = _get_sat_altitude(data_arr, alt_prefixes)
    except KeyError:
        if use_tle:
            logger.warning(
                    "Orbital parameters missing from metadata.  "
                    "Calculating from TLE using skyfield and astropy.")
            return _get_satpos_from_platform_name(data_arr)
        raise KeyError("Unable to determine satellite position. Either the "
                       "reader doesn't provide that information or "
                       "geolocation datasets were not available.")
    return lon, lat, alt


def _get_prefix_order_by_preference(prefixes, preference):
    preferred_prefixes = [prefix for prefix in prefixes if preference and preference in prefix]
    nonpreferred_prefixes = [prefix for prefix in prefixes if not preference or preference not in prefix]
    if nonpreferred_prefixes[-1] == "projection_":
        # remove projection as a prefix as it is our fallback
        nonpreferred_prefixes = nonpreferred_prefixes[:-1]
    return preferred_prefixes + nonpreferred_prefixes


def _get_sat_altitude(data_arr, key_prefixes):
    orb_params = data_arr.attrs["orbital_parameters"]
    alt_keys = [prefix + "altitude" for prefix in key_prefixes]
    try:
        alt = _get_first_available_item(orb_params, alt_keys)
    except KeyError:
        alt = orb_params["projection_altitude"]
        warnings.warn(
            "Actual satellite altitude not available, using projection altitude instead.",
            stacklevel=3
        )
    return alt


def _get_sat_lonlat(data_arr, key_prefixes):
    orb_params = data_arr.attrs["orbital_parameters"]
    lon_keys = [prefix + "longitude" for prefix in key_prefixes]
    lat_keys = [prefix + "latitude" for prefix in key_prefixes]
    try:
        lon = _get_first_available_item(orb_params, lon_keys)
        lat = _get_first_available_item(orb_params, lat_keys)
    except KeyError:
        lon = orb_params["projection_longitude"]
        lat = orb_params["projection_latitude"]
        warnings.warn(
            "Actual satellite lon/lat not available, using projection center instead.",
            stacklevel=3
        )
    return lon, lat


def _get_satpos_from_platform_name(cth_dataset):
    """Get satellite position if no orbital parameters in metadata.

    Some cloud top height datasets lack orbital parameter information in
    metadata.  Here, orbital parameters are calculated based on the platform
    name and start time, via Two Line Element (TLE) information.

    Needs pyorbital, skyfield, and astropy to be installed.
    """
    from pyorbital import tlefile
    from skyfield.api import EarthSatellite, load
    from skyfield.toposlib import wgs84

    name = cth_dataset.attrs["platform_name"]
    tle = tlefile.read(name)
    es = EarthSatellite(tle.line1, tle.line2, name)
    ts = load.timescale()
    gc = es.at(ts.from_datetime(
        cth_dataset.attrs["start_time"].replace(tzinfo=datetime.timezone.utc)))
    (lat, lon) = wgs84.latlon_of(gc)
    height = wgs84.height_of(gc).to("m")
    return (lon.degrees, lat.degrees, height.value)


def _get_first_available_item(data_dict, possible_keys):
    for possible_key in possible_keys:
        try:
            return data_dict[possible_key]
        except KeyError:
            continue
    raise KeyError("None of the possible keys found: {}".format(", ".join(possible_keys)))


def recursive_dict_update(d: MutableMapping, u: Mapping) -> None:
    """Recursive dictionary update.

    Copied from:

        http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth

    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = d.get(k, {})
            recursive_dict_update(r, v)
            d[k] = r
        else:
            d[k] = u[k]


def _check_yaml_configs(configs, key):
    """Get a diagnostic for the yaml *configs*.

    *key* is the section to look for to get a name for the config at hand.
    """
    diagnostic = {}
    for i in configs:
        for fname in i:
            msg = "ok"
            res = None
            with open(fname, "r", encoding="utf-8") as stream:
                try:
                    res = yaml.load(stream, Loader=UnsafeLoader)
                except yaml.YAMLError as err:
                    stream.seek(0)
                    res = yaml.load(stream, Loader=BaseLoader)
                    if err.context == "while constructing a Python object":
                        msg = err.problem
                    else:
                        msg = "error"
                finally:
                    try:
                        diagnostic[res[key]["name"]] = msg
                    except (KeyError, TypeError):
                        # this object doesn't have a 'name'
                        pass
    return diagnostic

def _check_package_version(package_name: str) -> Optional[str]:
    """Check the version of `package_name`.

    Args:
        package_name (str): the distribution package name.

    Returns:
        the version number if available else `None`.
    """
    try:
        return importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return None


def show_versions(packages=None):
    """Shows version for system, python and common packages (if installed).

    Args:
        packages (list or None): Limit packages to those specified.

    Returns:
        None.

    """
    packages = (
        (
            "cartopy",
            "geoviews",
            "numpy",
            "dask",
            "xarray",
            "gdal",
            "rasterio",
            "pyproj",
            "netcdf4",
            "h5py",
            "pyhdf",
            "h5netcdf",
            "fsspec",
        )
        if packages is None
        else packages
    )

    print("Versions")  # noqa: T201
    print("======")  # noqa: T201
    print(f"platform: {platform.platform()}")  # noqa: T201
    print(f"python: {platform.python_version()}")  # noqa: T201
    print()  # noqa: T201

    for package_name in sorted(packages):
        package_version = _check_package_version(package_name)
        print( # noqa: T201
            f"{package_name}: {package_version if package_version else 'not installed'}"
        )

    print()  # noqa: T201


def check_satpy(readers=None, writers=None, packages=None):
    """Check the satpy readers and writers for correct installation.

    Args:
        readers (list or None): Limit readers checked to those specified
        writers (list or None): Limit writers checked to those specified
        packages (list or None): Limit packages checked to those specified

    Returns:
        None

    """
    from satpy.readers.core.config import configs_for_reader
    from satpy.writers.core.config import configs_for_writer

    print("Readers")  # noqa: T201
    print("=======")  # noqa: T201
    for reader, res in sorted(_check_yaml_configs(configs_for_reader(reader=readers), "reader").items()):
        print(reader + ": ", res)  # noqa: T201
    print()  # noqa: T201

    print("Writers")  # noqa: T201
    print("=======")  # noqa: T201
    for writer, res in sorted(_check_yaml_configs(configs_for_writer(writer=writers), "writer").items()):
        print(writer + ": ", res)  # noqa: T201
    print()  # noqa: T201

    show_versions(packages=packages)


def unify_chunks(*data_arrays: xr.DataArray) -> tuple[xr.DataArray, ...]:
    """Run :func:`xarray.unify_chunks` if input dimensions are all the same size.

    This is mostly used in :class:`satpy.composites.core.CompositeBase` to safe
    guard against running :func:`dask.array.map_blocks` with arrays of
    different chunk sizes. Doing so can cause unexpected results or errors.
    However, xarray's ``unify_chunks`` will raise an exception if dimensions
    of the provided DataArrays are different sizes. This is a common case for
    Satpy. For example, the "bands" dimension may be 1 (L), 2 (LA), 3 (RGB), or
    4 (RGBA) for most compositor operations that combine other composites
    together.

    """
    if not hasattr(xr, "unify_chunks"):
        return data_arrays
    if not _all_dims_same_size(data_arrays):
        return data_arrays
    return tuple(xr.unify_chunks(*data_arrays))


def _all_dims_same_size(data_arrays: tuple[xr.DataArray, ...]) -> bool:
    known_sizes: dict[str, int] = {}
    for data_arr in data_arrays:
        for dim, dim_size in data_arr.sizes.items():
            known_size = known_sizes.setdefault(dim, dim_size)
            if dim_size != known_size:
                # this dimension is a different size than previously found
                # xarray.unify_chunks will error out if we tried to use it
                return False
    return True


@contextlib.contextmanager
def ignore_invalid_float_warnings():
    """Ignore warnings generated for working with NaN/inf values.

    Numpy and dask sometimes don't like NaN or inf values in normal function
    calls. This context manager hides/ignores them inside its context.

    Examples:
        Use around numpy operations that you expect to produce warnings::

            with ignore_invalid_float_warnings():
                np.nanmean(np.nan)

    """
    with np.errstate(invalid="ignore"), warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        yield


@contextlib.contextmanager
def ignore_pyproj_proj_warnings():
    """Wrap operations that we know will produce a PROJ.4 precision warning.

    Only to be used internally to Pyresample when we have no other choice but
    to use PROJ.4 strings/dicts. For example, serialization to YAML or other
    human-readable formats or testing the methods that produce the PROJ.4
    versions of the CRS.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            "You will likely lose important projection information",
            UserWarning,
        )
        yield


def get_chunk_size_limit(dtype=float):
    """Compute the chunk size limit in bytes given *dtype* (float by default).

    It is derived from PYTROLL_CHUNK_SIZE if defined (although deprecated) first, from dask config's `array.chunk-size`
    then. It defaults to 128MiB.

    Returns:
        The recommended chunk size in bytes.
    """
    pixel_size = _get_chunk_pixel_size()
    if pixel_size is not None:
        return pixel_size * np.dtype(dtype).itemsize
    return get_dask_chunk_size_in_bytes()


def get_dask_chunk_size_in_bytes():
    """Get the dask configured chunk size in bytes."""
    return dask.utils.parse_bytes(dask.config.get("array.chunk-size", "128MiB"))


def _get_chunk_pixel_size():
    """Compute the maximum chunk size from PYTROLL_CHUNK_SIZE."""
    legacy_chunk_size = _get_pytroll_chunk_size()
    if legacy_chunk_size is not None:
        return legacy_chunk_size ** 2


def get_legacy_chunk_size():
    """Get the legacy chunk size.

    This function should only be used while waiting for code to be migrated to use satpy.utils.get_chunk_size_limit
    instead.
    """
    chunk_size = _get_pytroll_chunk_size()

    if chunk_size is not None:
        return chunk_size

    import math

    return int(math.sqrt(get_dask_chunk_size_in_bytes() / 8))


def _get_pytroll_chunk_size():
    try:
        chunk_size = int(os.environ["PYTROLL_CHUNK_SIZE"])
        warnings.warn(
            "The PYTROLL_CHUNK_SIZE environment variable is pending deprecation. "
            "You can use the dask config setting `array.chunk-size` (or the DASK_ARRAY__CHUNK_SIZE environment"
            " variable) and set it to the square of the PYTROLL_CHUNK_SIZE instead.",
            stacklevel=2
        )
        return chunk_size
    except KeyError:
        return None


def normalize_low_res_chunks(
        chunks: tuple[int | Literal["auto"], ...],
        input_shape: tuple[int, ...],
        previous_chunks: tuple[int, ...],
        low_res_multipliers: tuple[int, ...],
        input_dtype: DTypeLike,
) -> tuple[int, ...]:
    """Compute dask chunk sizes based on data resolution.

    First, chunks are computed for the highest resolution version of the data.
    This is done by multiplying the input array shape by the
    ``low_res_multiplier`` and then using Dask's utility functions and
    configuration to produce a chunk size to fit into a specific number of
    bytes. See :doc:`dask:array-chunks` for more information.
    Next, the same multiplier is used to reduce the high resolution chunk sizes
    to the lower resolution of the input data. The end result of reading
    multiple resolutions of data is that each dask chunk covers the same
    geographic region. This also means replicating or aggregating one
    resolution and then combining arrays should not require any rechunking.

    Args:
        chunks: Requested chunk size for each dimension. This is passed
            directly to dask. Use ``"auto"`` for dimensions that should have
            chunks determined for them, ``-1`` for dimensions that should be
            whole (not chunked), and ``1`` or any other positive integer for
            dimensions that have a known chunk size beforehand.
        input_shape: Shape of the array to compute dask chunk size for.
        previous_chunks: Any previous chunking or structure of the data. This
            can also be thought of as the smallest number of high (fine) resolution
            elements that make up a single "unit" or chunk of data. This could
            be a multiple or factor of the scan size for some instruments and/or
            could be based on the on-disk chunk size. This value ensures that
            chunks are aligned to the underlying data structure for best
            performance. On-disk chunk sizes should be multiplied by the
            largest low resolution multiplier if it is the same between all
            files (ex. 500m file has 226 chunk size, 1km file has 226 chunk
            size, etc).. Otherwise, the resulting low resolution chunks may
            not be aligned to the on-disk chunks. For example, if dask decides
            on a chunk size of 226 * 3 for 500m data, that becomes 226 * 3 / 2
            for 1km data which is not aligned to the on-disk chunk size of 226.
        low_res_multipliers: Number of high (fine) resolution pixels that fit
            in a single low (coarse) resolution pixel.
        input_dtype: Dtype for the final unscaled array. This is usually
            32-bit float (``np.float32``) or 64-bit float (``np.float64``)
            for non-category data. If this doesn't represent the final data
            type of the data then the final size of chunks in memory will not
            match the user's request via dask's ``array.chunk-size``
            configuration. Sometimes it is useful to keep this as a single
            dtype for all reading functionality (ex. ``np.float32``) in order
            to keep all read variable chunks the same size regardless of dtype.

    Returns:
        A tuple where each element is the chunk size for that axis/dimension.

    """
    if any(len(input_shape) != len(param) for param in (low_res_multipliers, chunks, previous_chunks)):
        raise ValueError("Input shape, low res multipliers, chunks, and previous chunks must all be the same size")
    high_res_shape = tuple(dim_size * lr_mult for dim_size, lr_mult in zip(input_shape, low_res_multipliers))
    chunks_for_high_res = dask.array.core.normalize_chunks(
        chunks,
        shape=high_res_shape,
        dtype=input_dtype,
        previous_chunks=previous_chunks,
    )
    low_res_chunks: list[int] = []
    for req_chunks, hr_chunks, prev_chunks, lr_mult in zip(
            chunks,
            chunks_for_high_res,
            previous_chunks, low_res_multipliers
    ):
        if req_chunks != "auto":
            low_res_chunks.append(req_chunks)
            continue
        low_res_chunks.append(round(max(hr_chunks[0] / lr_mult, prev_chunks / lr_mult)))
    return tuple(low_res_chunks)


def convert_remote_files_to_fsspec(filenames, storage_options=None):
    """Check filenames for transfer protocols, convert to FSFile objects if possible."""
    if storage_options is None:
        storage_options = {}
    if isinstance(filenames, dict):
        return _check_file_protocols_for_dicts(filenames, storage_options)
    return _check_file_protocols(filenames, storage_options)


def _check_file_protocols_for_dicts(filenames, storage_options):
    res = {}
    for reader, files in filenames.items():
        opts = storage_options.get(reader, {})
        res[reader] = _check_file_protocols(files, opts)
    return res


def _check_file_protocols(filenames, storage_options):
    local_files, remote_files, fs_files = _sort_files_to_local_remote_and_fsfiles(filenames)

    if remote_files:
        return local_files + fs_files + _filenames_to_fsfile(remote_files, storage_options)

    return local_files + fs_files


def _sort_files_to_local_remote_and_fsfiles(filenames):
    from satpy.readers.core.remote import FSFile

    local_files = []
    remote_files = []
    fs_files = []
    for f in filenames:
        if isinstance(f, FSFile):
            fs_files.append(f)
        elif isinstance(f, pathlib.Path):
            local_files.append(f)
        elif urlparse(f).scheme in ("", "file") or "\\" in f:
            local_files.append(f)
        else:
            remote_files.append(f)
    return local_files, remote_files, fs_files


def _filenames_to_fsfile(filenames, storage_options):
    import fsspec

    from satpy.readers.core.remote import FSFile

    if filenames:
        fsspec_files = fsspec.open_files(filenames, **storage_options)
        return [FSFile(f) for f in fsspec_files]
    return []


def get_storage_options_from_reader_kwargs(reader_kwargs):
    """Read and clean storage options from reader_kwargs."""
    if reader_kwargs is None:
        return None, None
    new_reader_kwargs = deepcopy(reader_kwargs)  # don't modify user provided dict
    storage_options = _get_storage_dictionary_options(new_reader_kwargs)
    return storage_options, new_reader_kwargs


def _get_storage_dictionary_options(reader_kwargs):
    storage_opt_dict = {}
    shared_storage_options = reader_kwargs.pop("storage_options", {})
    if not reader_kwargs:
        # no other reader kwargs
        return shared_storage_options
    for reader_name, rkwargs in reader_kwargs.items():
        if not isinstance(rkwargs, dict):
            # reader kwargs are not per-reader, return a single dictionary of storage options
            return shared_storage_options
        if shared_storage_options:
            # set base storage options if there are any
            storage_opt_dict[reader_name] = shared_storage_options.copy()
        if isinstance(rkwargs, dict) and "storage_options" in rkwargs:
            storage_opt_dict.setdefault(reader_name, {}).update(rkwargs.pop("storage_options"))
    return storage_opt_dict


@contextmanager
def import_error_helper(dependency_name):
    """Give more info on an import error."""
    try:
        yield
    except ImportError as err:
        raise ImportError(err.msg + f" It can be installed with the {dependency_name} package.")


def find_in_ancillary(data, dataset):
    """Find a dataset by name in the ancillary vars of another dataset.

    Args:
        data (xarray.DataArray):
            Array for which to search the ancillary variables
        dataset (str):
            Name of ancillary variable to look for.
    """
    matches = [x for x in data.attrs["ancillary_variables"] if x.attrs.get("name") == dataset]
    cnt = len(matches)
    if cnt < 1:
        raise ValueError(
            f"Could not find dataset named {dataset:s} in ancillary "
            f"variables for dataset {data.attrs.get('name')!r}")
    if cnt > 1:
        raise ValueError(
            f"Expected exactly one dataset named {dataset:s} in ancillary "
            f"variables for dataset {data.attrs.get('name')!r}, "
            f"found {cnt:d}")
    return matches[0]


def datetime64_to_pydatetime(dt64: np.datetime64) -> datetime.datetime:
    """Convert numpy.datetime64 timestamp to Python datetime.

    Discards nanosecond precision, because Python datetime only has microsecond
    precision.

    Args:
        dt64: Timestamp to be converted
    Returns:
        Converted timestamp
    """
    return dt64.astype("datetime64[us]").astype(datetime.datetime)


def _import_and_warn_new_location(new_module: str, name: str) -> Any:
    import inspect
    from importlib import import_module

    frame = inspect.stack()[1]
    old_module_name = inspect.getmodule(frame[0]).__name__  # type: ignore

    mod = import_module(new_module)
    obj = getattr(mod, name)

    warnings.warn(
        f"'{old_module_name}.{name}' has been moved to '{new_module}.{name}'. "
        f"Import from the new location instead (ex. 'from {new_module} import {name}'). "
        "The old import paths will be removed in Satpy 1.0",
        stacklevel=3,
    )
    return obj


def flatten_dict(d, parent_key="", sep="_"):
    """Flatten a nested dictionary.

    Based on https://stackoverflow.com/a/6027615/5703449
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, parent_key=new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Simple command line too that converts an image into a NinJo Tiff file.

NinJo Tiff metadata can be passed as command line input or
through a config file (an example is given in the ninjo-cmd.yaml
file in the etc directory).

The area matching the input image shall be defined in the
areas configuration file (located in $PPP_CONFIG_DIR).

"""

import argparse
import os

import yaml
from yaml import UnsafeLoader

from satpy import Scene
from satpy.area import get_area_def
from satpy.utils import debug_on

debug_on()

parser = argparse.ArgumentParser(description="Turn an image into a NinjoTiff.")
parser.add_argument("--cfg", dest="cfg", action="store",
                    help="YAML configuration as an alternative to the command line input for NinJo metadata.")
parser.add_argument("--input_dir", dest="input_dir", action="store",
                    help="Directory with input data, that must contain a timestamp in the filename.")
parser.add_argument("--chan_id", dest="chan_id", action="store", help="Channel ID", default="9999")
parser.add_argument("--sat_id", dest="sat_id", action="store", help="Satellite ID", default="8888")
parser.add_argument("--data_cat", dest="data_cat", action="store",
                    help="Category of data (one of GORN, GPRN, PORN)", default="GORN")
parser.add_argument("--area", dest="areadef", action="store",
                    help="Area name, the definition must exist in your areas configuration file",
                    default="nrEURO1km_NPOL_COALeqc")
parser.add_argument("--ph_unit", dest="ph_unit", action="store", help="Physical unit", default="CELSIUS")
parser.add_argument("--data_src", dest="data_src", action="store", help="Data source", default="EUMETCAST")
args = parser.parse_args()

if (args.input_dir is not None):
    os.chdir(args.input_dir)

cfg = vars(args)
if (args.cfg is not None):
    with open(args.cfg, "r") as ymlfile:
        cfg = yaml.load(ymlfile, Loader=UnsafeLoader)

narea = get_area_def(args.areadef)
global_data = Scene(reader="generic_image")
global_data.load(["image"])

global_data["image"].info["area"] = narea
fname = global_data["image"].info["filename"]
ofname = fname[:-3] + "tif"

# global_data.save_dataset('image', filename="out.png", writer="simple_image")
global_data.save_dataset("image", filename=ofname, writer="ninjotiff",
                         sat_id=cfg["sat_id"],
                         chan_id=cfg["chan_id"],
                         data_cat=cfg["data_cat"],
                         data_source=cfg["data_src"],
                         physic_unit=cfg["ph_unit"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019.
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""A reader for HDF5 Snow Cover (SC) file produced by the Hydrology SAF."""

import datetime as dt
import logging

import dask.array as da
import h5py
import numpy as np
import xarray as xr

from satpy.area import get_area_def
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

LOG = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()
AREA_X_OFFSET = 1211
AREA_Y_OFFSET = 62


class HSAFFileHandler(BaseFileHandler):
    """File handler for HSAF H5 files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(HSAFFileHandler, self).__init__(filename,
                                              filename_info,
                                              filetype_info)
        self._h5fh = h5py.File(self.filename, "r")

    @property
    def end_time(self):
        """Get end time."""
        return self.start_time + dt.timedelta(hours=23, minutes=59, seconds=59)

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["sensing_time"]

    def _prepare_variable_for_palette(self, dset, ds_info):
        colormap = np.array(dset)
        return xr.DataArray(colormap, attrs=ds_info, dims=("idx", "RGB"))

    def get_metadata(self, dset, name):
        """Get the metadata."""
        ds_info = {"name": name}
        if name == "SC":
            ds_info.update({
                "filename": self.filename,
                "data_time": self.start_time,
                "nx": dset.shape[1],
                "ny": dset.shape[0]
            })
        return ds_info

    def get_area_def(self, dsid):
        """Area definition for h10 SC dataset.

        Since it is not available in the HDF5 message,
        using hardcoded one (it's known).
        """
        if dsid["name"] == "SC":
            return self._get_area_def()
        raise NotImplementedError

    def _get_area_def(self):
        """Area definition for h10 - hardcoded.

        Area definition not available in the HDF5 message,
        so using hardcoded one (it's known).

        ::

            hsaf_h10:
              description: H SAF H10 area definition
              projection:
                proj: geos
                lon_0: 0
                h: 35785831
                x_0: 0
                y_0: 0
                a: 6378169
                rf: 295.488065897001
                no_defs: null
                type: crs
              shape:
                height: 916
                width: 1902
              area_extent:
                lower_left_xy: [-1936760.3163240477, 2635854.280233425]
                upper_right_xy: [3770006.7195370505, 5384223.683413638]
                units: m

        """
        fd_def = get_area_def("msg_seviri_fes_3km")
        hsaf_def = fd_def[AREA_Y_OFFSET:AREA_Y_OFFSET+916,
                          AREA_X_OFFSET:AREA_X_OFFSET+1902]

        return hsaf_def

    def _get_dataset(self, ds_name):
        if ds_name == "SC_pal":
            _ds_name = "colormap"
        else:
            _ds_name = ds_name
        return self._h5fh.get(_ds_name)

    def get_dataset(self, ds_id, ds_info):
        """Read a HDF5 file into an xarray DataArray."""
        ds = self._get_dataset(ds_id["name"])
        ds_info = self.get_metadata(ds, ds_id["name"])

        if ds_id["name"] == "SC":
            ds_info["start_time"] = self.start_time
            ds_info["data_time"] = self.start_time
            ds_info["end_time"] = self.end_time

            data = da.from_array(ds, chunks=CHUNK_SIZE)
            return xr.DataArray(data, attrs=ds_info, dims=("y", "x"))

        elif ds_id["name"] == "SC_pal":
            return self._prepare_variable_for_palette(ds, ds_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""SAFE SAR-C reader.

This module implements a reader for Sentinel 1 SAR-C GRD (level1) SAFE format as
provided by ESA. The format is comprised of a directory containing multiple
files, most notably two measurement files in geotiff and a few xml files for
calibration, noise and metadata.

References:
  - *Level 1 Product Formatting*
    https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1-product-formatting

  - J. Park, A. A. Korosov, M. Babiker, S. Sandven and J. Won,
    *"Efficient Thermal Noise Removal for Sentinel-1 TOPSAR Cross-Polarization Channel,"*
    in IEEE Transactions on Geoscience and Remote Sensing, vol. 56, no. 3,
    pp. 1555-1565, March 2018.
    doi: `10.1109/TGRS.2017.2765248 <https://doi.org/10.1109/TGRS.2017.2765248>`_

"""

import functools
import json
import logging
import warnings
from collections import defaultdict
from datetime import timezone as tz
from functools import cached_property
from pathlib import Path
from threading import Lock

import defusedxml.ElementTree as ET
import numpy as np
import rasterio
import rioxarray  # noqa F401  # xarray open_dataset use engine rasterio, which use rioxarray
import xarray as xr
from dask import array as da
from geotiepoints.geointerpolator import lonlat2xyz, xyz2lonlat
from geotiepoints.interpolator import MultipleSplineInterpolator
from xarray import DataArray

from satpy.dataset.data_dict import DatasetDict
from satpy.dataset.dataid import DataID
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.readers.core.yaml_reader import GenericYAMLReader
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


def dictify(r):
    """Convert an ElementTree into a dict."""
    return {r.tag: _dictify(r)}


def _dictify(r):
    """Convert an xml element to dict."""
    d = {}
    if r.text and r.text.strip():
        try:
            return int(r.text)
        except ValueError:
            try:
                return np.float32(r.text)
            except ValueError:
                return r.text
    for x in r.findall("./*"):
        if x.tag in d and not isinstance(d[x.tag], list):
            d[x.tag] = [d[x.tag]]
            d[x.tag].append(_dictify(x))
        else:
            d[x.tag] = _dictify(x)
    return d


def _get_calibration_name(calibration):
    """Get the proper calibration name."""
    calibration_name = getattr(calibration, "name", calibration) or "gamma"
    if calibration_name == "sigma_nought":
        calibration_name = "sigmaNought"
    elif calibration_name == "beta_nought":
        calibration_name = "betaNought"
    return calibration_name


class SAFEXML(BaseFileHandler):
    """XML file reader for the SAFE format."""

    def __init__(self, filename, filename_info, filetype_info,
                 header_file=None, image_shape=None):
        """Init the xml filehandler."""
        super().__init__(filename, filename_info, filetype_info)

        self._start_time = filename_info["start_time"].replace(tzinfo=tz.utc)
        self._end_time = filename_info["end_time"].replace(tzinfo=tz.utc)
        self._polarization = filename_info["polarization"]
        if isinstance(self.filename, str):
            self.filename = Path(self.filename)
        with self.filename.open() as fd:
            self.root = ET.parse(fd)
        self._image_shape = image_shape

    def get_metadata(self):
        """Convert the xml metadata to dict."""
        return dictify(self.root.getroot())

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time


class SAFEXMLAnnotation(SAFEXML):
    """XML file reader for the SAFE format, Annotation file."""

    def __init__(self, filename, filename_info, filetype_info,
                 header_file=None):
        """Init the XML annotation reader."""
        super().__init__(filename, filename_info, filetype_info, header_file)
        self.get_incidence_angle = functools.lru_cache(maxsize=10)(
            self._get_incidence_angle_uncached
        )
        self.hdr = self.get_metadata()
        self._image_shape = (self.hdr["product"]["imageAnnotation"]["imageInformation"]["numberOfLines"],
                             self.hdr["product"]["imageAnnotation"]["imageInformation"]["numberOfSamples"])

    @property
    def image_shape(self):
        """Return the image shape of this dataset."""
        return self._image_shape

    def get_dataset(self, key, info, chunks=None):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return

        if key["name"] == "incidence_angle":
            return self.get_incidence_angle(chunks=chunks or CHUNK_SIZE)

    def _get_incidence_angle_uncached(self, chunks):
        """Get the incidence angle array."""
        incidence_angle = XMLArray(self.root, ".//geolocationGridPoint", "incidenceAngle")
        return incidence_angle.expand(self._image_shape, chunks=chunks)


class Calibrator(SAFEXML):
    """XML file reader for the SAFE format, Calibration file."""

    def __init__(self, filename, filename_info, filetype_info,
                 header_file=None, image_shape=None):
        """Init the XML calibration reader."""
        super().__init__(filename, filename_info, filetype_info, header_file, image_shape)
        self.get_calibration = functools.lru_cache(maxsize=10)(
            self._get_calibration_uncached
        )

    def get_dataset(self, key, info, chunks=None):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return
        if key["name"] == "calibration_constant":
            return self.get_calibration_constant()
        return self.get_calibration(key["name"], chunks=chunks or CHUNK_SIZE)

    def get_calibration_constant(self):
        """Load the calibration constant."""
        return np.float32(self.root.find(".//absoluteCalibrationConstant").text)

    def _get_calibration_uncached(self, calibration, chunks=None):
        """Get the calibration array."""
        calibration_name = _get_calibration_name(calibration)
        calibration_vector = self._get_calibration_vector(calibration_name, chunks)
        return calibration_vector

    def _get_calibration_vector(self, calibration_name, chunks):
        """Get the calibration vector."""
        calibration_vector = XMLArray(self.root, ".//calibrationVector", calibration_name)
        return calibration_vector.expand(self._image_shape, chunks=chunks)

    def __call__(self, dn, calibration_type, chunks=None):
        """Calibrate the data."""
        logger.debug("Reading calibration data.")
        cal = self.get_calibration(calibration_type, chunks=chunks)
        cal_constant = self.get_calibration_constant()
        logger.debug("Calibrating.")
        data = ((dn + cal_constant) / (cal ** 2)).clip(min=0)
        return data

class Denoiser(SAFEXML):
    """XML file reader for the SAFE format, Noise file."""

    def __init__(self, filename, filename_info, filetype_info,
                 header_file=None, image_shape=None):
        """Init the xml filehandler."""
        super().__init__(filename, filename_info, filetype_info, header_file, image_shape)

        self.azimuth_noise_reader = AzimuthNoiseReader(self.root, self._image_shape)
        self.get_noise_correction = functools.lru_cache(maxsize=10)(
            self._get_noise_correction_uncached
        )

    def get_dataset(self, key, info, chunks=None):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return
        if key["name"] == "noise":
            return self.get_noise_correction(chunks=chunks or CHUNK_SIZE)

    def _get_noise_correction_uncached(self, chunks=None):
        """Get the noise correction array."""
        try:
            noise = self.read_legacy_noise(chunks)
        except KeyError:
            range_noise = self.read_range_noise_array(chunks)
            azimuth_noise = self.azimuth_noise_reader.read_azimuth_noise_array(chunks)
            noise = range_noise * azimuth_noise
        return noise

    def read_legacy_noise(self, chunks):
        """Read noise for legacy GRD data."""
        noise = XMLArray(self.root, ".//noiseVector", "noiseLut")
        return noise.expand(self._image_shape, chunks)

    def read_range_noise_array(self, chunks):
        """Read the range-noise array."""
        range_noise = XMLArray(self.root, ".//noiseRangeVector", "noiseRangeLut")
        return range_noise.expand(self._image_shape, chunks)

    def __call__(self, dn, chunks):
        """Denoise the data."""
        logger.debug("Reading noise data.")
        noise = self.get_noise_correction(chunks=chunks).fillna(0)
        dn = dn - noise
        return dn



class AzimuthNoiseReader:
    """Class to parse and read azimuth-noise data.

    The azimuth noise vector is provided as a series of blocks, each comprised
    of a column of data to fill the block and a start and finish column number,
    and a start and finish line.
    For example, we can see here a (fake) azimuth noise array::

        [[ 1.  1.  1. nan nan nan nan nan nan nan]
         [ 1.  1.  1. nan nan nan nan nan nan nan]
         [ 2.  2.  3.  3.  3.  4.  4.  4.  4. nan]
         [ 2.  2.  3.  3.  3.  4.  4.  4.  4. nan]
         [ 2.  2.  3.  3.  3.  4.  4.  4.  4. nan]
         [ 2.  2.  5.  5.  5.  5.  6.  6.  6.  6.]
         [ 2.  2.  5.  5.  5.  5.  6.  6.  6.  6.]
         [ 2.  2.  5.  5.  5.  5.  6.  6.  6.  6.]
         [ 2.  2.  7.  7.  7.  7.  7.  8.  8.  8.]
         [ 2.  2.  7.  7.  7.  7.  7.  8.  8.  8.]]

    As is shown here, the blocks may not cover the full array, and hence it has
    to be gap-filled with NaNs.
    """

    def __init__(self, root, shape):
        """Set up the azimuth noise reader."""
        self.root = root
        self.elements = self.root.findall(".//noiseAzimuthVector")
        self._image_shape = shape
        self.blocks = []

    def read_azimuth_noise_array(self, chunks=CHUNK_SIZE):
        """Read the azimuth noise vectors."""
        self._read_azimuth_noise_blocks(chunks)
        populated_array = self._assemble_azimuth_noise_blocks(chunks)

        return populated_array

    def _read_azimuth_noise_blocks(self, chunks):
        """Read the azimuth noise blocks."""
        self.blocks = []
        for elt in self.elements:
            block = _AzimuthBlock(elt)
            new_arr = block.expand(chunks)
            self.blocks.append(new_arr)

    def _assemble_azimuth_noise_blocks(self, chunks):
        """Assemble the azimuth noise blocks into one single array."""
        # The strategy here is a bit convoluted. The job would be trivial if
        # performed on regular numpy arrays, but here we want to keep the data
        # as xarray/dask array as much as possible.
        # Using a pure xarray approach was tested (with `combine_first`,
        # `interpolate_na`, etc), but was found to be memory-hungry at the time
        # of implementation (March 2021). Hence the usage of a custom algorithm,
        # relying mostly on dask arrays.
        slices = self._create_dask_slices_from_blocks(chunks)
        populated_array = da.vstack(slices).rechunk(chunks)
        populated_array = xr.DataArray(populated_array, dims=["y", "x"],
                                       coords={"x": np.arange(self._image_shape[1]),
                                               "y": np.arange(self._image_shape[0])})
        return populated_array

    def _create_dask_slices_from_blocks(self, chunks):
        """Create full-width slices from azimuth noise blocks."""
        current_line = 0
        slices = []
        while current_line < self._image_shape[0]:
            new_slice = self._create_dask_slice_from_block_line(current_line, chunks)
            slices.append(new_slice)
            current_line += new_slice.shape[0]
        return slices

    def _create_dask_slice_from_block_line(self, current_line, chunks):
        """Create a dask slice from the blocks at the current line."""
        pieces = self._get_array_pieces_for_current_line(current_line)
        dask_pieces = self._get_padded_dask_pieces(pieces, chunks)
        new_slice = da.hstack(dask_pieces)

        return new_slice

    def _get_array_pieces_for_current_line(self, current_line):
        """Get the array pieces that cover the current line."""
        current_blocks = self._find_blocks_covering_line(current_line)
        current_blocks.sort(key=(lambda x: x.coords["x"][0]))
        next_line = self._get_next_start_line(current_blocks, current_line)
        current_y = np.arange(current_line, next_line, dtype=np.uint16)
        pieces = [arr.sel(y=current_y) for arr in current_blocks]
        return pieces

    def _find_blocks_covering_line(self, current_line):
        """Find the blocks covering a given line."""
        current_blocks = []
        for block in self.blocks:
            if block.coords["y"][0] <= current_line <= block.coords["y"][-1]:
                current_blocks.append(block)
        return current_blocks

    def _get_next_start_line(self, current_blocks, current_line):
        next_line = min((arr.coords["y"][-1] for arr in current_blocks)) + 1
        blocks_starting_soon = [block for block in self.blocks if current_line < block.coords["y"][0] < next_line]
        if blocks_starting_soon:
            next_start_line = min((arr.coords["y"][0] for arr in blocks_starting_soon))
            next_line = min(next_line, next_start_line)
        return next_line

    def _get_padded_dask_pieces(self, pieces, chunks):
        """Get the padded pieces of a slice."""
        pieces = sorted(pieces, key=(lambda x: x.coords["x"][0]))
        dask_pieces = []
        previous_x_end = -1
        piece = pieces[0]
        next_x_start = piece.coords["x"][0].item()
        y_shape = len(piece.coords["y"])

        x_shape = (next_x_start - previous_x_end - 1)
        self._fill_dask_pieces(dask_pieces, (y_shape, x_shape), chunks)

        for i, piece in enumerate(pieces):
            dask_pieces.append(piece.data)
            previous_x_end = piece.coords["x"][-1].item()
            try:
                next_x_start = pieces[i + 1].coords["x"][0].item()
            except IndexError:
                next_x_start = self._image_shape[1]

            x_shape = (next_x_start - previous_x_end - 1)
            self._fill_dask_pieces(dask_pieces, (y_shape, x_shape), chunks)

        return dask_pieces

    @staticmethod
    def _fill_dask_pieces(dask_pieces, shape, chunks):
        if shape[1] > 0:
            new_piece = da.full(shape, np.nan, chunks=chunks, dtype=np.float32)
            dask_pieces.append(new_piece)


class _AzimuthBlock:
    """Implementation of an single azimuth-noise block."""

    def __init__(self, xml_element):
        """Set up the block from an XML element."""
        self.element = xml_element

    def expand(self, chunks):
        """Build an azimuth block from xml data."""
        corr = 1
        # This isn't needed with newer data (> 2020). When was the change operated?
        #
        #         The azimuth noise is normalized per swath to account for gain
        #         differences between the swaths in EW mode.
        #
        #         This is based on the this reference:
        #         J. Park, A. A. Korosov, M. Babiker, S. Sandven and J. Won,
        #         "Efficient Thermal Noise Removal for Sentinel-1 TOPSAR Cross-Polarization Channel,"
        #         in IEEE Transactions on Geoscience and Remote Sensing, vol. 56, no. 3,
        #         pp. 1555-1565, March 2018.
        #         doi: 10.1109/TGRS.2017.2765248
        #
        # For old data. < 2020
        # swath = elt.find('swath').text
        # if swath == 'EW1':
        #     corr = 1.5
        # if swath in ['EW4', 'IW3']:
        #     corr = 1.2
        # if swath == 'EW5':
        #     corr = 1.5
        data = self.lut * corr

        x_coord = np.arange(self.first_pixel, self.last_pixel + 1, dtype=np.uint16)
        y_coord = np.arange(self.first_line, self.last_line + 1, dtype=np.uint16)
        new_arr = (da.ones((len(y_coord), len(x_coord)), dtype=np.float32, chunks=chunks) *
                   np.interp(y_coord, self.lines, data)[:, np.newaxis].astype(np.float32))
        new_arr = xr.DataArray(new_arr,
                               dims=["y", "x"],
                               coords={"x": x_coord,
                                       "y": y_coord})
        return new_arr

    @property
    def first_pixel(self):
        return np.uint16(self.element.find("firstRangeSample").text)

    @property
    def last_pixel(self):
        return np.uint16(self.element.find("lastRangeSample").text)

    @property
    def first_line(self):
        return np.uint16(self.element.find("firstAzimuthLine").text)

    @property
    def last_line(self):
        return np.uint16(self.element.find("lastAzimuthLine").text)

    @property
    def lines(self):
        lines = self.element.find("line").text.split()
        return np.array(lines).astype(np.uint16)

    @property
    def lut(self):
        lut = self.element.find("noiseAzimuthLut").text.split()
        return np.array(lut, dtype=np.float32)


class XMLArray:
    """A proxy for getting xml data as an array."""

    def __init__(self, root, list_tag, element_tag):
        """Set up the XML array."""
        self.root = root
        self.list_tag = list_tag
        self.element_tag = element_tag
        self.data, self.low_res_coords = self._read_xml_array()

    def expand(self, shape, chunks=None):
        """Generate the full-blown array."""
        return self.interpolate_xml_array(shape, chunks=chunks)

    def _read_xml_array(self):
        """Read an array from xml."""
        elements = self.get_data_items()
        y = []
        x = []
        data = []
        for elt in elements:
            new_x = elt.find("pixel").text.split()
            y += [int(elt.find("line").text)] * len(new_x)
            x += [int(val) for val in new_x]
            data += [np.float32(val)
                     for val in elt.find(self.element_tag).text.split()]

        return np.asarray(data), (x, y)

    def get_data_items(self):
        """Get the data items for this array."""
        data_items = self.root.findall(self.list_tag)
        if not data_items:
            raise KeyError("Can't find data items for xml tag " + self.list_tag)
        return data_items

    def interpolate_xml_array(self, shape, chunks):
        """Interpolate arbitrary size dataset to a full sized grid."""
        xpoints, ypoints = self.low_res_coords
        return interpolate_xarray_linear(xpoints, ypoints, self.data, shape, chunks=chunks)


def intp(grid_x, grid_y, interpolator):
    """Interpolate."""
    return interpolator((grid_y, grid_x))


def interpolate_xarray_linear(xpoints, ypoints, values, shape, chunks=CHUNK_SIZE):
    """Interpolate linearly, generating a dask array."""
    from scipy.interpolate import LinearNDInterpolator

    if isinstance(chunks, (list, tuple)):
        vchunks, hchunks = chunks
    else:
        vchunks, hchunks = chunks, chunks
    points = np.vstack((np.asarray(ypoints, dtype=np.uint16),
                        np.asarray(xpoints, dtype=np.uint16))).T

    interpolator = LinearNDInterpolator(points, values)

    grid_x, grid_y = da.meshgrid(da.arange(shape[1], chunks=hchunks, dtype=np.uint16),
                                 da.arange(shape[0], chunks=vchunks, dtype=np.uint16))

    # workaround for non-thread-safe first call of the interpolator:
    interpolator((0, 0))
    res = da.map_blocks(intp, grid_x, grid_y, interpolator=interpolator).astype(values.dtype)

    return DataArray(res, dims=("y", "x"))


class SAFEGRD(BaseFileHandler):
    """Measurement file reader.

    The measurement files are in geotiff format and read using rasterio. For
    performance reasons, the reading adapts the chunk size to match the file's
    block size.
    """

    def __init__(self, filename, filename_info, filetype_info, calibrator, denoiser):
        """Init the grd filehandler."""
        super().__init__(filename, filename_info, filetype_info)
        self._start_time = filename_info["start_time"].replace(tzinfo=tz.utc)
        self._end_time = filename_info["end_time"].replace(tzinfo=tz.utc)

        self._polarization = filename_info["polarization"]

        self._mission_id = filename_info["mission_id"]

        self.calibrator = calibrator
        self.denoiser = denoiser
        self.read_lock = Lock()

        self.get_lonlatalts = functools.lru_cache(maxsize=2)(
            self._get_lonlatalts_uncached
        )

    def get_dataset(self, key, info):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return

        logger.debug("Reading %s.", key["name"])

        if key["name"] in ["longitude", "latitude", "altitude"]:
            logger.debug("Constructing coordinate arrays.")
            arrays = dict()
            arrays["longitude"], arrays["latitude"], arrays["altitude"] = self.get_lonlatalts()

            data = arrays[key["name"]]
            data.attrs.update(info)

        else:
            data = self._calibrate_and_denoise(self._data, key)
            data.attrs.update(info)
            data.attrs.update({"platform_name": self._mission_id})

            data = self._change_quantity(data, key["quantity"])

        return data

    @cached_property
    def _data(self):
        data = xr.open_dataarray(open_file_or_filename(self.filename, mode="rb"), engine="rasterio",
                                 chunks="auto"
                                ).squeeze()
        self.chunks = data.data.chunksize
        data = data.assign_coords(x=np.arange(len(data.coords["x"])),
                                      y=np.arange(len(data.coords["y"])))

        return data

    @staticmethod
    def _change_quantity(data, quantity):
        """Change quantity to dB if needed."""
        if quantity == "dB":
            data.data = 10 * np.log10(data.data)
            data.attrs["units"] = "dB"
        else:
            data.attrs["units"] = "1"

        return data

    def _calibrate_and_denoise(self, data, key):
        """Calibrate and denoise the data."""
        dn = self._get_digital_number(data)
        dn = self.denoiser(dn, self.chunks)
        data = self.calibrator(dn, key["calibration"], self.chunks)

        return data

    def _get_digital_number(self, data):
        """Get the digital numbers (uncalibrated data)."""
        data = data.where(data > 0)
        data = data.astype(np.float32)
        dn = data * data
        return dn

    def _get_lonlatalts_uncached(self):
        """Obtain GCPs and construct latitude and longitude arrays.

        Returns:
           (tuple): A tuple with longitude and latitude arrays
        """
        shape = self._data.shape

        (xpoints, ypoints), (gcp_lons, gcp_lats, gcp_alts), (gcps, crs) = self.get_gcps()

        fine_points = [np.arange(size) for size in shape]
        x, y, z = lonlat2xyz(gcp_lons, gcp_lats)


        interpolator = MultipleSplineInterpolator((ypoints, xpoints), x, y, z, gcp_alts, kx=2, ky=2)
        hx, hy, hz, altitudes = interpolator.interpolate(fine_points, chunks=self.chunks)


        longitudes, latitudes = xyz2lonlat(hx, hy, hz)
        altitudes = xr.DataArray(altitudes, dims=["y", "x"])
        longitudes = xr.DataArray(longitudes, dims=["y", "x"])
        latitudes = xr.DataArray(latitudes, dims=["y", "x"])

        longitudes.attrs["gcps"] = gcps
        longitudes.attrs["crs"] = crs
        latitudes.attrs["gcps"] = gcps
        latitudes.attrs["crs"] = crs
        altitudes.attrs["gcps"] = gcps
        altitudes.attrs["crs"] = crs

        return longitudes, latitudes, altitudes

    def get_gcps(self):
        """Read GCP from the GDAL band.

        Returns:
           (tuple) Pixel and Line indices 1d arrays
           (tuple) longitude and latitude 1d arrays
           (tuple) ground control points and crs

        """
        gcps = get_gcps_from_array(self._data)
        crs = self._data.rio.crs

        gcp_list = [(feature["properties"]["row"], feature["properties"]["col"], *feature["geometry"]["coordinates"])
                    for feature in gcps["features"]]
        gcp_array = np.array(gcp_list)

        ypoints = np.unique(gcp_array[:, 0]).astype(np.uint16)
        xpoints = np.unique(gcp_array[:, 1]).astype(np.uint16)

        gcp_lons = gcp_array[:, 2].reshape(ypoints.shape[0], xpoints.shape[0])
        gcp_lats = gcp_array[:, 3].reshape(ypoints.shape[0], xpoints.shape[0])
        gcp_alts = gcp_array[:, 4].reshape(ypoints.shape[0], xpoints.shape[0])

        rio_gcps = [rasterio.control.GroundControlPoint(*gcp) for gcp in gcp_list]

        return (xpoints, ypoints), (gcp_lons, gcp_lats, gcp_alts), (rio_gcps, crs)

    def get_bounding_box(self):
        """Get the bounding box for the data coverage."""
        (xpoints, ypoints), (gcp_lons, gcp_lats, gcp_alts), (rio_gcps, crs) = self.get_gcps()
        bblons = np.hstack((gcp_lons[0, :-1], gcp_lons[:-1, -1], gcp_lons[-1, :1:-1], gcp_lons[:1:-1, 0]))
        bblats = np.hstack((gcp_lats[0, :-1], gcp_lats[:-1, -1], gcp_lats[-1, :1:-1], gcp_lats[:1:-1, 0]))
        return bblons.tolist(), bblats.tolist()

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time


class SAFESARReader(GenericYAMLReader):
    """A reader for SAFE SAR-C data for Sentinel 1 satellites."""

    def __init__(self, config, filter_parameters=None):
        """Set up the SAR reader."""
        super().__init__(config)
        self.filter_parameters = filter_parameters
        self.files_by_type = defaultdict(list)
        self.storage_items = []

    @property
    def start_time(self):
        """Get the start time."""
        return self.storage_items.values()[0].filename_info["start_time"].replace(tzinfo=tz.utc)

    @property
    def end_time(self):
        """Get the end time."""
        return self.storage_items.values()[0].filename_info["end_time"].replace(tzinfo=tz.utc)

    def load(self, dataset_keys, **kwargs):
        """Load some data."""
        if kwargs:
            warnings.warn(f"Don't know how to handle kwargs {kwargs}")
        datasets = DatasetDict()
        for key in dataset_keys:
            for handler in self.storage_items.values():
                val = handler.get_dataset(key, info=dict())
                if val is not None:
                    val.attrs["start_time"] = handler.start_time
                    if key["name"] not in ["longitude", "latitude"]:
                        lonlats = self.load([DataID(self._id_keys, name="longitude", polarization=key["polarization"]),
                                             DataID(self._id_keys, name="latitude", polarization=key["polarization"])])
                        gcps = get_gcps_from_array(val)
                        from pyresample.future.geometry import SwathDefinition
                        val.attrs["area"] = SwathDefinition(lonlats["longitude"], lonlats["latitude"],
                                                            attrs=dict(gcps=gcps,
                                                                       bounding_box=handler.get_bounding_box()))
                    datasets[key] = val
                    continue
        return datasets

    def create_storage_items(self, files, **kwargs):
        """Create the storage items."""
        self.files_by_type = self._get_files_by_type(files)
        image_shapes = self._get_image_shapes()
        calibrators = self._create_calibrators(image_shapes)
        denoisers = self._create_denoisers(image_shapes)
        measurement_handlers = self._create_measurement_handlers(calibrators, denoisers)

        self.storage_items = measurement_handlers


    def _get_files_by_type(self, files):
        files_by_type = defaultdict(list)
        for file_type, type_info in self.config["file_types"].items():
            files_by_type[file_type].extend(self.filename_items_for_filetype(files, type_info))
        return files_by_type


    def _get_image_shapes(self):
        image_shapes = dict()
        for annotation_file, annotation_info in self.files_by_type["safe_annotation"]:
            annotation_fh = SAFEXMLAnnotation(annotation_file,
                                              filename_info=annotation_info,
                                              filetype_info=None)
            image_shapes[annotation_info["polarization"]] = annotation_fh.image_shape
        return image_shapes


    def _create_calibrators(self, image_shapes):
        calibrators = dict()
        for calibration_file, calibration_info in self.files_by_type["safe_calibration"]:
            polarization = calibration_info["polarization"]
            calibrators[polarization] = Calibrator(calibration_file,
                                                   filename_info=calibration_info,
                                                   filetype_info=None,
                                                   image_shape=image_shapes[polarization])

        return calibrators


    def _create_denoisers(self, image_shapes):
        denoisers = dict()
        for noise_file, noise_info in self.files_by_type["safe_noise"]:
            polarization = noise_info["polarization"]
            denoisers[polarization] = Denoiser(noise_file,
                                               filename_info=noise_info,
                                               filetype_info=None,
                                               image_shape=image_shapes[polarization])

        return denoisers


    def _create_measurement_handlers(self, calibrators, denoisers):
        measurement_handlers = dict()
        for measurement_file, measurement_info in self.files_by_type["safe_measurement"]:
            polarization = measurement_info["polarization"]
            measurement_handlers[polarization] = SAFEGRD(measurement_file,
                                                         filename_info=measurement_info,
                                                         calibrator=calibrators[polarization],
                                                         denoiser=denoisers[polarization],
                                                         filetype_info=None)

        return measurement_handlers


def get_gcps_from_array(val):
    """Get the gcps from the spatial_ref coordinate as a geojson dict."""
    gcps = val.coords["spatial_ref"].attrs["gcps"]
    if isinstance(gcps, str):
        gcps = json.loads(gcps)
    return gcps



#!/usr/bin/env python
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for NWC SAF MSG (2013) reader."""

import os
import tempfile
import unittest
from collections import OrderedDict

import h5py
import numpy as np
import pytest

from satpy.tests.reader_tests.utils import fill_h5
from satpy.tests.utils import RANDOM_GEN

CTYPE_TEST_ARRAY = (RANDOM_GEN.random((1856, 3712)) * 255).astype(np.uint8)
CTYPE_TEST_FRAME = (np.arange(100).reshape(10, 10) / 100. * 20).astype(np.uint8)
CTYPE_TEST_ARRAY[1000:1010, 1000:1010] = CTYPE_TEST_FRAME

CTTH_HEIGHT_TEST_ARRAY = (RANDOM_GEN.random((1856, 3712)) * 255).astype(np.uint8)
_CTTH_HEIGHT_TEST_FRAME = (np.arange(100).reshape(10, 10) / 100. * 80).astype(np.uint8)
CTTH_HEIGHT_TEST_ARRAY[1000:1010, 1000:1010] = _CTTH_HEIGHT_TEST_FRAME

CTTH_HEIGHT_TEST_FRAME_RES = _CTTH_HEIGHT_TEST_FRAME.astype(np.float32) * 200 - 2000
CTTH_HEIGHT_TEST_FRAME_RES[0, 0:10] = np.nan
CTTH_HEIGHT_TEST_FRAME_RES[1, 0:3] = np.nan

CTTH_PRESSURE_TEST_ARRAY = (RANDOM_GEN.random((1856, 3712)) * 255).astype(np.uint8)
_CTTH_PRESSURE_TEST_FRAME = (np.arange(100).reshape(10, 10) / 100. * 54).astype(np.uint8)
CTTH_PRESSURE_TEST_ARRAY[1000:1010, 1000:1010] = _CTTH_PRESSURE_TEST_FRAME

CTTH_PRESSURE_TEST_FRAME_RES = _CTTH_PRESSURE_TEST_FRAME.astype(np.float32) * 25 - 250
CTTH_PRESSURE_TEST_FRAME_RES[0, 0:10] = np.nan
CTTH_PRESSURE_TEST_FRAME_RES[1, 0:9] = np.nan

CTTH_TEMPERATURE_TEST_ARRAY = (RANDOM_GEN.random((1856, 3712)) * 255).astype(np.uint8)
_CTTH_TEMPERATURE_TEST_FRAME = (np.arange(100).reshape(10, 10) / 100. * 140).astype(np.uint8)
_CTTH_TEMPERATURE_TEST_FRAME[8, 5] = 255
CTTH_TEMPERATURE_TEST_ARRAY[1000:1010, 1000:1010] = _CTTH_TEMPERATURE_TEST_FRAME

CTTH_TEMPERATURE_TEST_FRAME_RES = _CTTH_TEMPERATURE_TEST_FRAME.astype(np.float32) * 1.0 + 150
CTTH_TEMPERATURE_TEST_FRAME_RES[8, 5] = np.nan


fake_ct = {
    "01-PALETTE": {
        "attrs": {
            "CLASS": b"PALETTE",
            "PAL_COLORMODEL": b"RGB",
            "PAL_TYPE": b"DIRECTINDEX",
        },
        "value": np.array(
            [
                [100, 100, 100],
                [0, 120, 0],
                [0, 0, 0],
                [250, 190, 250],
                [220, 160, 220],
                [255, 150, 0],
                [255, 100, 0],
                [255, 220, 0],
                [255, 180, 0],
                [255, 255, 140],
                [240, 240, 0],
                [250, 240, 200],
                [215, 215, 150],
                [255, 255, 255],
                [230, 230, 230],
                [0, 80, 215],
                [0, 180, 230],
                [0, 240, 240],
                [90, 200, 160],
                [200, 0, 200],
                [95, 60, 30],
            ],
            dtype=np.uint8,
        ),
    },
    "02-PALETTE": {
        "attrs": {
            "CLASS": b"PALETTE",
            "PAL_COLORMODEL": b"RGB",
            "PAL_TYPE": b"DIRECTINDEX",
        },
        "value": np.array(
            [[100, 100, 100], [255, 100, 0], [0, 80, 215], [95, 60, 30]], dtype=np.uint8
        ),
    },
    "CT": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CT",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": 0.0,
            "PALETTE": "<HDF5 object reference> 01-PALETTE",
            "PRODUCT": b"CT__",
            "SCALING_FACTOR": 1.0,
        },
        "value": (CTYPE_TEST_ARRAY),
    },
    "CT_PHASE": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CT_PHASE",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": 0.0,
            "PALETTE": "<HDF5 object reference> 02-PALETTE",
            "PRODUCT": b"CT__",
            "SCALING_FACTOR": 1.0,
        },
        "value": (RANDOM_GEN.random((1856, 3712)) * 255).astype(np.uint8),
    },
    "CT_QUALITY": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CT_QUALITY",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": 0.0,
            "PRODUCT": b"CT__",
            "SCALING_FACTOR": 1.0,
        },
        "value": (RANDOM_GEN.random((1856, 3712)) * 65535).astype(np.uint16),
    },
    "attrs": {
        "CFAC": 13642337,
        "COFF": 1856,
        "GP_SC_ID": 323,
        "IMAGE_ACQUISITION_TIME": b"201611090800",
        "LFAC": 13642337,
        "LOFF": 1856,
        "NB_PARAMETERS": 3,
        "NC": 3712,
        "NL": 1856,
        "NOMINAL_PRODUCT_TIME": b"201611090814",
        "PACKAGE": b"SAFNWC/MSG",
        "PRODUCT_ALGORITHM_VERSION": b"             2.2",
        "PRODUCT_NAME": b"CT__",
        "PROJECTION_NAME": b"GEOS<+000.0>",
        "REGION_NAME": b"MSG-N",
        "SAF": b"NWC",
        "SGS_PRODUCT_COMPLETENESS": 99,
        "SGS_PRODUCT_QUALITY": 79,
        "SPECTRAL_CHANNEL_ID": 0,
    },
}

fake_ct = OrderedDict(sorted(fake_ct.items(), key=lambda t: t[0]))

fake_ctth = {
    "01-PALETTE": {
        "attrs": {
            "CLASS": b"PALETTE",
            "PAL_COLORMODEL": b"RGB",
            "PAL_TYPE": b"DIRECTINDEX",
        },
        "value": np.array(
            [
                [0, 0, 0],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [255, 255, 255],
                [239, 239, 223],
                [239, 239, 223],
                [238, 214, 210],
                [238, 214, 210],
                [0, 255, 255],
                [0, 255, 255],
                [0, 216, 255],
                [0, 216, 255],
                [0, 178, 255],
                [0, 178, 255],
                [0, 140, 48],
                [0, 140, 48],
                [0, 255, 0],
                [0, 255, 0],
                [153, 255, 0],
                [153, 255, 0],
                [178, 255, 0],
                [178, 255, 0],
                [216, 255, 0],
                [216, 255, 0],
                [255, 255, 0],
                [255, 255, 0],
                [255, 216, 0],
                [255, 216, 0],
                [255, 164, 0],
                [255, 164, 0],
                [255, 102, 0],
                [255, 102, 0],
                [255, 76, 0],
                [255, 76, 0],
                [178, 51, 0],
                [178, 51, 0],
                [153, 20, 47],
                [153, 20, 47],
                [126, 0, 43],
                [126, 0, 43],
                [255, 0, 216],
                [255, 0, 216],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
                [255, 0, 128],
            ],
            dtype=np.uint8,
        ),
    },
    "02-PALETTE": {
        "attrs": {
            "CLASS": b"PALETTE",
            "PAL_COLORMODEL": b"RGB",
            "PAL_TYPE": b"DIRECTINDEX",
        },
        "value": (RANDOM_GEN.random((128, 3)) * 255).astype(np.uint8),
    },
    "03-PALETTE": {
        "attrs": {
            "CLASS": b"PALETTE",
            "PAL_COLORMODEL": b"RGB",
            "PAL_TYPE": b"DIRECTINDEX",
        },
        "value": (RANDOM_GEN.random((256, 3)) * 255).astype(np.uint8),
    },
    "04-PALETTE": {
        "attrs": {
            "CLASS": b"PALETTE",
            "PAL_COLORMODEL": b"RGB",
            "PAL_TYPE": b"DIRECTINDEX",
        },
        "value": np.array(
            [
                [78, 119, 145],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [0, 0, 0],
                [12, 12, 12],
                [24, 24, 24],
                [36, 36, 36],
                [48, 48, 48],
                [60, 60, 60],
                [72, 72, 72],
                [84, 84, 84],
                [96, 96, 96],
                [108, 108, 108],
                [120, 120, 120],
                [132, 132, 132],
                [144, 144, 144],
                [156, 156, 156],
                [168, 168, 168],
                [180, 180, 180],
                [192, 192, 192],
                [204, 204, 204],
                [216, 216, 216],
                [228, 228, 228],
                [240, 240, 240],
                [240, 240, 240],
            ],
            dtype=np.uint8,
        ),
    },
    "CTTH_EFFECT": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CTTH_EFFECT",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": -50.0,
            "PALETTE": "<HDF5 object reference> 04-PALETTE",
            "PRODUCT": b"CTTH",
            "SCALING_FACTOR": 5.0,
        },
        "value": (RANDOM_GEN.random((1856, 3712)) * 255).astype(np.uint8),
    },
    "CTTH_HEIGHT": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CTTH_HEIGHT",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": -2000.0,
            "PALETTE": "<HDF5 object reference> 02-PALETTE",
            "PRODUCT": b"CTTH",
            "SCALING_FACTOR": 200.0,
        },
        "value": (CTTH_HEIGHT_TEST_ARRAY),
    },
    "CTTH_PRESS": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CTTH_PRESS",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": -250.0,
            "PALETTE": "<HDF5 object reference> 01-PALETTE",
            "PRODUCT": b"CTTH",
            "SCALING_FACTOR": 25.0,
        },
        "value": (CTTH_PRESSURE_TEST_ARRAY),
    },
    "CTTH_QUALITY": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CTTH_QUALITY",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": 0.0,
            "PRODUCT": b"CTTH",
            "SCALING_FACTOR": 1.0,
        },
        "value": (RANDOM_GEN.random((1856, 3712)) * 65535).astype(np.uint16),
    },
    "CTTH_TEMPER": {
        "attrs": {
            "CLASS": b"IMAGE",
            "ID": b"CTTH_TEMPER",
            "IMAGE_COLORMODEL": b"RGB",
            "IMAGE_SUBCLASS": b"IMAGE_INDEXED",
            "IMAGE_VERSION": b"1.0",
            "N_COLS": 3712,
            "N_LINES": 1856,
            "OFFSET": 150.0,
            "PALETTE": "<HDF5 object reference> 03-PALETTE",
            "PRODUCT": b"CTTH",
            "SCALING_FACTOR": 1.0,
        },
        "value": (CTTH_TEMPERATURE_TEST_ARRAY),
    },
    "attrs": {
        "CFAC": 13642337,
        "COFF": 1856,
        "GP_SC_ID": 323,
        "IMAGE_ACQUISITION_TIME": b"201611090800",
        "LFAC": 13642337,
        "LOFF": 1856,
        "NB_PARAMETERS": 5,
        "NC": 3712,
        "NL": 1856,
        "NOMINAL_PRODUCT_TIME": b"201611090816",
        "PACKAGE": b"SAFNWC/MSG",
        "PRODUCT_ALGORITHM_VERSION": b"             2.2",
        "PRODUCT_NAME": b"CTTH",
        "PROJECTION_NAME": b"GEOS<+000.0>",
        "REGION_NAME": b"MSG-N",
        "SAF": b"NWC",
        "SGS_PRODUCT_COMPLETENESS": 87,
        "SGS_PRODUCT_QUALITY": 69,
        "SPECTRAL_CHANNEL_ID": 0,
    },
}

fake_ctth = OrderedDict(sorted(fake_ctth.items(), key=lambda t: t[0]))

PROJ_KM = {
    "gdal_projection": "+proj=geos +a=6378.137000 +b=6356.752300 +lon_0=0.000000 +h=35785.863000",
    "gdal_xgeo_up_left": -5569500.0,
    "gdal_ygeo_up_left": 5437500.0,
    "gdal_xgeo_low_right": 5566500.0,
    "gdal_ygeo_low_right": 2653500.0,
}
PROJ = {
    "gdal_projection": "+proj=geos +a=6378137.000 +b=6356752.300 +lon_0=0.000000 +h=35785863.000",
    "gdal_xgeo_up_left": -5569500.0,
    "gdal_ygeo_up_left": 5437500.0,
    "gdal_xgeo_low_right": 5566500.0,
    "gdal_ygeo_low_right": 2653500.0,
}

AREA_DEF_DICT = {
    "proj_dict": {"proj": "geos", "lon_0": 0, "h": 35785831, "x_0": 0, "y_0": 0,
                  "a": 6378169, "b": 6356583.8, "units": "m", "no_defs": None, "type": "crs"},
    "area_id": "MSG-N",
    "x_size": 3712,
    "y_size": 1856,
    "area_extent": (-5570248.2825, 1501.0099, 5567247.8793, 5570247.8784)
}


class TestH5NWCSAF(unittest.TestCase):
    """Test the nwcsaf msg reader."""

    def setUp(self):
        """Set up the tests."""
        self.filename_ct = os.path.join(
            tempfile.gettempdir(),
            "SAFNWC_MSG3_CT___201611090800_MSG-N_______.PLAX.CTTH.0.h5",
        )

        self.filename_ctth = os.path.join(
            tempfile.gettempdir(),
            "SAFNWC_MSG3_CTTH_201611090800_MSG-N_______.PLAX.CTTH.0.h5",
        )

        def cut_h5_object_ref(root, attr):
            if isinstance(attr, str) and attr.startswith("<HDF5 object reference>"):
                return root[attr[24:]].ref
            return attr

        h5f = h5py.File(self.filename_ct, mode="w")
        fill_h5(h5f, fake_ct, attr_processor=cut_h5_object_ref)
        for attr, val in fake_ct["attrs"].items():
            h5f.attrs[attr] = val
        h5f.close()

        h5f = h5py.File(self.filename_ctth, mode="w")
        fill_h5(h5f, fake_ctth, attr_processor=cut_h5_object_ref)
        for attr, val in fake_ctth["attrs"].items():
            h5f.attrs[attr] = val
        h5f.close()

    def test_get_area_def(self):
        """Get the area definition."""
        from pyproj import CRS

        from satpy.readers.nwcsaf_msg2013_hdf5 import Hdf5NWCSAF
        from satpy.tests.utils import make_dataid

        filename_info = {}
        filetype_info = {}
        dsid = make_dataid(name="ct")
        test = Hdf5NWCSAF(self.filename_ct, filename_info, filetype_info)

        area_def = test.get_area_def(dsid)

        aext_res = AREA_DEF_DICT["area_extent"]
        for i in range(4):
            assert area_def.area_extent[i] == pytest.approx(aext_res[i], abs=1e-4)

        expected_crs = CRS(AREA_DEF_DICT["proj_dict"])
        assert expected_crs == area_def.crs

        assert AREA_DEF_DICT["x_size"] == area_def.width
        assert AREA_DEF_DICT["y_size"] == area_def.height

        assert AREA_DEF_DICT["area_id"] == area_def.area_id

    def test_get_dataset(self):
        """Retrieve datasets from a NWCSAF msgv2013 hdf5 file."""
        from satpy.readers.nwcsaf_msg2013_hdf5 import Hdf5NWCSAF
        from satpy.tests.utils import make_dataid

        filename_info = {}
        filetype_info = {}
        dsid = make_dataid(name="ct")
        test = Hdf5NWCSAF(self.filename_ct, filename_info, filetype_info)
        ds = test.get_dataset(dsid, {"file_key": "CT"})
        assert ds.shape == (1856, 3712)
        assert ds.dtype == np.uint8
        np.testing.assert_allclose(ds.data[1000:1010, 1000:1010].compute(), CTYPE_TEST_FRAME)

        filename_info = {}
        filetype_info = {}
        dsid = make_dataid(name="ctth_alti")
        test = Hdf5NWCSAF(self.filename_ctth, filename_info, filetype_info)
        ds = test.get_dataset(dsid, {"file_key": "CTTH_HEIGHT"})
        assert ds.shape == (1856, 3712)
        assert ds.dtype == np.float32
        np.testing.assert_allclose(ds.data[1000:1010, 1000:1010].compute(), CTTH_HEIGHT_TEST_FRAME_RES)

        filename_info = {}
        filetype_info = {}
        dsid = make_dataid(name="ctth_pres")
        test = Hdf5NWCSAF(self.filename_ctth, filename_info, filetype_info)
        ds = test.get_dataset(dsid, {"file_key": "CTTH_PRESS"})
        assert ds.shape == (1856, 3712)
        assert ds.dtype == np.float32
        np.testing.assert_allclose(ds.data[1000:1010, 1000:1010].compute(), CTTH_PRESSURE_TEST_FRAME_RES)

        filename_info = {}
        filetype_info = {}
        dsid = make_dataid(name="ctth_tempe")
        test = Hdf5NWCSAF(self.filename_ctth, filename_info, filetype_info)
        ds = test.get_dataset(dsid, {"file_key": "CTTH_TEMPER"})
        assert ds.shape == (1856, 3712)
        assert ds.dtype == np.float32
        np.testing.assert_allclose(ds.data[1000:1010, 1000:1010].compute(), CTTH_TEMPERATURE_TEST_FRAME_RES)

    def tearDown(self):
        """Destroy."""
        try:
            os.remove(self.filename_ct)
            os.remove(self.filename_ctth)
        except OSError:
            pass



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Shared objects and base classes for writers."""
from __future__ import annotations

import logging
import os
import typing
import warnings

from satpy.aux_download import DataDownloadMixin
from satpy.plugin_base import Plugin
from satpy.writers.core.compute import compute_writer_results, split_results

if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from os import PathLike
    from typing import Any

    import dask.array as da
    import xarray as xr
    from dask.delayed import Delayed

LOG = logging.getLogger(__name__)


class Writer(Plugin, DataDownloadMixin):
    """Base Writer class for all other writers.

    A minimal writer subclass should implement the `save_dataset` method.
    """

    def __init__(self, name=None, filename=None, base_dir=None, **kwargs):
        """Initialize the writer object.

        Args:
            name (str): A name for this writer for log and error messages.
                If this writer is configured in a YAML file its name should
                match the name of the YAML file. Writer names may also appear
                in output file attributes.
            filename (str): Filename to save data to. This filename can and
                should specify certain python string formatting fields to
                differentiate between data written to the files. Any
                attributes provided by the ``.attrs`` of a DataArray object
                may be included. Format and conversion specifiers provided by
                the :class:`trollsift <trollsift.parser.StringFormatter>`
                package may also be used. Any directories in the provided
                pattern will be created if they do not exist. Example::

                    {platform_name}_{sensor}_{name}_{start_time:%Y%m%d_%H%M%S}.tif

            base_dir (str):
                Base destination directories for all created files.
            kwargs (dict): Additional keyword arguments to pass to the
                :class:`~satpy.plugin_base.Plugin` class.

        """
        # Load the config
        Plugin.__init__(self, **kwargs)
        self.info = self.config.get("writer", {})

        if "file_pattern" in self.info:
            warnings.warn(
                "Writer YAML config is using 'file_pattern' which "
                "has been deprecated, use 'filename' instead.",
                stacklevel=2
            )
            self.info["filename"] = self.info.pop("file_pattern")

        if "file_pattern" in kwargs:
            warnings.warn(
                "'file_pattern' has been deprecated, use 'filename' instead.",
                DeprecationWarning,
                stacklevel=2
            )
            filename = kwargs.pop("file_pattern")

        # Use options from the config file if they weren't passed as arguments
        self.name = self.info.get("name", None) if name is None else name
        self.file_pattern = self.info.get("filename", None) if filename is None else filename

        if self.name is None:
            raise ValueError("Writer 'name' not provided")

        self.filename_parser = self.create_filename_parser(base_dir)
        self.register_data_files()

    @classmethod
    def separate_init_kwargs(cls, kwargs):
        """Help separating arguments between init and save methods.

        Currently the :class:`~satpy.scene.Scene` is passed one set of
        arguments to represent the Writer creation and saving steps. This is
        not preferred for Writer structure, but provides a simpler interface
        to users. This method splits the provided keyword arguments between
        those needed for initialization and those needed for the ``save_dataset``
        and ``save_datasets`` method calls.

        Writer subclasses should try to prefer keyword arguments only for the
        save methods only and leave the init keyword arguments to the base
        classes when possible.

        """
        # FUTURE: Don't pass Scene.save_datasets kwargs to init and here
        init_kwargs = {}
        kwargs = kwargs.copy()
        for kw in ["base_dir", "filename", "file_pattern"]:
            if kw in kwargs:
                init_kwargs[kw] = kwargs.pop(kw)
        return init_kwargs, kwargs

    def create_filename_parser(self, base_dir):
        """Create a :class:`trollsift.parser.Parser` object for later use."""
        from trollsift import parser

        # just in case a writer needs more complex file patterns
        # Set a way to create filenames if we were given a pattern
        if base_dir and self.file_pattern:
            file_pattern = os.path.join(base_dir, self.file_pattern)
        else:
            file_pattern = self.file_pattern
        return parser.Parser(file_pattern) if file_pattern else None

    @staticmethod
    def _prepare_metadata_for_filename_formatting(attrs):
        if isinstance(attrs.get("sensor"), set):
            attrs["sensor"] = "-".join(sorted(attrs["sensor"]))

    def get_filename(self, **kwargs):
        """Create a filename where output data will be saved.

        Args:
            kwargs (dict): Attributes and other metadata to use for formatting
                the previously provided `filename`.

        """
        if self.filename_parser is None:
            raise RuntimeError("No filename pattern or specific filename provided")
        self._prepare_metadata_for_filename_formatting(kwargs)
        output_filename = self.filename_parser.compose(kwargs)
        dirname = os.path.dirname(output_filename)
        if dirname and not os.path.isdir(dirname):
            LOG.info("Creating output directory: {}".format(dirname))
            os.makedirs(dirname, exist_ok=True)
        return output_filename

    def save_datasets(
            self,
            datasets: Iterable[xr.DataArray],
            compute: bool = True,
            **kwargs,
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[typing.Any]] | list[str | PathLike | None]:
        """Save all datasets to one or more files.

        Subclasses can use this method to save all datasets to one single
        file or optimize the writing of individual datasets. By default
        this simply calls `save_dataset` for each dataset provided.

        Args:
            datasets: Iterable of `xarray.DataArray` objects to save using
                this writer.
            compute: If `True` (default), compute all the saves to
                disk. If `False` then the return value is either
                a :doc:`dask:delayed` object or two lists to
                be passed to a :func:`dask.array.store` call.
                See return values below for more details.
            **kwargs: Keyword arguments to pass to `save_dataset`. See that
                documentation for more details.

        Returns:
            Value returned depends on `compute` keyword argument. If
            `compute` is `True` the value is the result of either a
            :func:`dask.array.store` operation or a :doc:`dask:delayed`
            compute, typically this is `None`. If `compute` is `False` then
            the result is either a :doc:`dask:delayed` object that can be
            computed with `delayed.compute()` or a two element tuple of
            sources and targets to be passed to :func:`dask.array.store`. If
            `targets` is provided then it is the caller's responsibility to
            close any objects that have a "close" method.

        """
        sources = []
        targets = []
        delayeds = []
        for ds in datasets:
            res = self.save_dataset(ds, compute=False, **kwargs)
            s, t, d = split_results([res])
            sources.extend(s)
            targets.extend(t)
            delayeds.extend(d)

        if compute:
            LOG.info("Computing and writing results...")
            to_compute = (sources, targets) if sources else delayeds
            return compute_writer_results([to_compute])

        if sources:
            return (sources, targets)
        return delayeds

    def save_dataset(
            self,
            dataset: xr.DataArray,
            filename: str | None = None,
            fill_value: float | int | None = None,
            compute: bool = True,
            units: str | None = None,
            **kwargs,
    ) -> list[da.Array | Delayed] | tuple[list[da.Array], list[Any]] | list[str | PathLike | None]:
        """Save the ``dataset`` to a given ``filename``.

        This method must be overloaded by the subclass.

        Args:
            dataset: Dataset to save using this writer.
            filename: Optionally specify the filename to save this
                dataset to. If not provided then `filename`
                which can be provided to the init method will be
                used and formatted by dataset attributes.
            fill_value: Replace invalid values in the dataset with this fill
                value if applicable to this writer.
            compute: If `True` (default), compute and save the dataset.
                If `False` return either a :doc:`dask:delayed`
                object or tuple of (source, target). See the
                return values below for more information.
            units: If not None, will convert the dataset to
                the given unit using pint-xarray before
                saving. Default is not to do any
                conversion.
            **kwargs: Other keyword arguments for this particular writer.

        Returns:
            Value returned depends on `compute`. If `compute` is `True` then
            the return value is the result of computing a
            :doc:`dask:delayed` object or running :func:`dask.array.store`.
            If `compute` is `False` then the returned value is either a
            :doc:`dask:delayed` object that can be computed using
            `delayed.compute()` or a tuple of (source, target) that should be
            passed to :func:`dask.array.store`. If target is provided the
            caller is responsible for calling `target.close()` if the target
            has this method.

        """
        raise NotImplementedError(
            "Writer '%s' has not implemented dataset saving" % (self.name, ))
